import pytest
from lbox.exceptions import (
    LabelboxError,
    ResourceNotFoundError,
)

from labelbox.schema.labeling_service import LabelingServiceStatus


def test_start_labeling_service(project):
    # Check that the labeling service is now available
    labeling_service = project.get_labeling_service()  # creates and gets it
    assert labeling_service.status == LabelingServiceStatus.SetUp
    assert labeling_service.project_id == project.uid

    labeling_service_status = project.get_labeling_service_status()
    assert labeling_service_status == LabelingServiceStatus.SetUp


def test_request_labeling_service_moe_offline_project(
    rand_gen,
    offline_chat_evaluation_project,
    chat_evaluation_ontology,
    offline_conversational_data_row,
    model_config,
):
    project = offline_chat_evaluation_project
    project.connect_ontology(chat_evaluation_ontology)

    project.create_batch(
        rand_gen(str),
        [offline_conversational_data_row.uid],  # sample of data row objects
    )

    project.upsert_instructions("tests/integration/media/sample_pdf.pdf")

    labeling_service = project.get_labeling_service()
    labeling_service.request()
    assert (
        project.get_labeling_service_status() == LabelingServiceStatus.Requested
    )


def test_request_labeling_service_moe_project(
    live_chat_evaluation_project_with_batch,
    chat_evaluation_ontology,
    model_config,
):
    project = live_chat_evaluation_project_with_batch
    project.connect_ontology(chat_evaluation_ontology)

    project.upsert_instructions("tests/integration/media/sample_pdf.pdf")

    labeling_service = project.get_labeling_service()
    with pytest.raises(
        LabelboxError,
        match='[{"errorType":"PROJECT_MODEL_CONFIG","errorMessage":"Project model config is not completed"}]',
    ):
        labeling_service.request()
    project.add_model_config(model_config.uid)
    project.set_project_model_setup_complete()

    labeling_service.request()
    assert (
        project.get_labeling_service_status() == LabelingServiceStatus.Requested
    )


def test_request_labeling_service_incomplete_requirements(ontology, project):
    labeling_service = (
        project.get_labeling_service()
    )  # project fixture is an Image type project
    with pytest.raises(
        ResourceNotFoundError, match="Associated ontology id could not be found"
    ):  # No labeling service by default
        labeling_service.request()
    project.connect_ontology(ontology)
    with pytest.raises(LabelboxError):
        labeling_service.request()



from collections import defaultdict
from typing import Any, Callable, Dict, List, Union, Optional, get_args

import labelbox
from labelbox.data.annotation_types.data import GenericDataRowData, MaskData

from ...annotated_types import Cuid
from .annotation import ClassificationAnnotation, ObjectAnnotation
from .relationship import RelationshipAnnotation
from .llm_prompt_response.prompt import PromptClassificationAnnotation
from .classification import ClassificationAnswer
from .geometry import Mask
from .metrics import ScalarMetric, ConfusionMatrixMetric
from .video import VideoClassificationAnnotation
from .video import VideoObjectAnnotation, VideoMaskAnnotation
from .mmc import MessageEvaluationTaskAnnotation
from pydantic import BaseModel, field_validator


class Label(BaseModel):
    """Container for holding data and annotations

    >>> Label(
    >>>    data = {'global_key': 'my-data-row-key'} # also accepts uid, external_id as keys
    >>>    annotations = [
    >>>        ObjectAnnotation(
    >>>            value = Point(x = 10, y = 10),
    >>>            name = "target"
    >>>        )
    >>>     ]
    >>>  )

    Args:
        uid: Optional Label Id in Labelbox
        data: GenericDataRowData or dict with a single key uid | global_key | external_id.
        annotations: List of Annotations in the label
        extra: additional context
    """

    uid: Optional[Cuid] = None
    data: Union[GenericDataRowData, MaskData]
    annotations: List[
        Union[
            ClassificationAnnotation,
            ObjectAnnotation,
            VideoMaskAnnotation,
            ScalarMetric,
            ConfusionMatrixMetric,
            RelationshipAnnotation,
            PromptClassificationAnnotation,
            MessageEvaluationTaskAnnotation,
        ]
    ] = []
    extra: Dict[str, Any] = {}
    is_benchmark_reference: Optional[bool] = False

    @field_validator("data", mode="before")
    def validate_data(cls, data):
        if isinstance(data, Dict):
            return GenericDataRowData(**data)
        return data

    def object_annotations(self) -> List[ObjectAnnotation]:
        return self._get_annotations_by_type(ObjectAnnotation)

    def classification_annotations(self) -> List[ClassificationAnnotation]:
        return self._get_annotations_by_type(ClassificationAnnotation)

    def _get_annotations_by_type(self, annotation_type):
        return [
            annot
            for annot in self.annotations
            if isinstance(annot, annotation_type)
        ]

    def frame_annotations(
        self,
    ) -> Dict[str, Union[VideoObjectAnnotation, VideoClassificationAnnotation]]:
        frame_dict = defaultdict(list)
        for annotation in self.annotations:
            if isinstance(
                annotation,
                (VideoObjectAnnotation, VideoClassificationAnnotation),
            ):
                frame_dict[annotation.frame].append(annotation)
        return frame_dict

    def add_url_to_masks(self, signer) -> "Label":
        """
        Creates signed urls for all masks in the Label.
        Multiple masks can reference the same MaskData mask so this makes sure we only upload that url once.
        Only uploads url if one doesn't already exist.

        Args:
            signer: A function that accepts bytes and returns a signed url.
        Returns:
            Label with updated references to new mask url
        """
        masks = []
        for annotation in self.annotations:
            # Allows us to upload shared masks once
            if isinstance(annotation.value, Mask):
                in_list = False
                for mask in masks:
                    if annotation.value.mask is mask:
                        in_list = True
                if not in_list:
                    masks.append(annotation.value.mask)
        for mask in masks:
            mask.create_url(signer)
        return self

    def create_data_row(
        self, dataset: "labelbox.Dataset", signer: Callable[[bytes], str]
    ) -> "Label":
        """
        Creates a data row and adds to the given dataset.
        Updates the label's data object to have the same external_id and uid as the data row.

        Args:
            dataset: labelbox dataset object to add the new data row to
            signer: A function that accepts bytes and returns a signed url.
        Returns:
            Label with updated references to new data row
        """
        args = {"row_data": self.data.create_url(signer)}
        if self.data.external_id is not None:
            args.update({"external_id": self.data.external_id})

        if self.data.uid is None:
            data_row = dataset.create_data_row(**args)
            self.data.uid = data_row.uid
            self.data.external_id = data_row.external_id
        return self

    def _assign_or_raise(self, annotation, lookup: Dict[str, str]) -> None:
        if annotation.feature_schema_id is not None:
            return

        feature_schema_id = lookup.get(annotation.name)
        if feature_schema_id is None:
            raise ValueError(
                f"No tool matches name {annotation.name}. "
                f"Must be one of {list(lookup.keys())}."
            )
        annotation.feature_schema_id = feature_schema_id

    def _assign_option(
        self, classification: ClassificationAnnotation, lookup: Dict[str, str]
    ) -> None:
        if isinstance(classification.value.answer, str):
            pass
        elif isinstance(classification.value.answer, ClassificationAnswer):
            self._assign_or_raise(classification.value.answer, lookup)
        elif isinstance(classification.value.answer, list):
            for answer in classification.value.answer:
                self._assign_or_raise(answer, lookup)
        else:
            raise TypeError(
                f"Unexpected type for answer found. {type(classification.value.answer)}"
            )

    @field_validator("annotations", mode="before")
    def validate_union(cls, value):
        supported = tuple(
            [
                field
                for field in get_args(
                    get_args(cls.model_fields["annotations"].annotation)[0]
                )
            ]
        )
        if not isinstance(value, list):
            raise TypeError(f"Annotations must be a list. Found {type(value)}")
        prompt_count = 0
        for v in value:
            if not isinstance(v, supported):
                raise TypeError(
                    f"Annotations should be a list containing the following classes : {supported}. Found {type(v)}"
                )
            # Validates only one prompt annotation is included
            if isinstance(v, PromptClassificationAnnotation):
                prompt_count += 1
                if prompt_count > 1:
                    raise TypeError(
                        f"Only one prompt annotation is allowed per label"
                    )
        return value



import pytest
from lbox.exceptions import LabelboxError, ResourceNotFoundError

import labelbox as lb
from labelbox.schema.foundry.app import App
from labelbox.schema.foundry.foundry_client import FoundryClient

# Yolo object detection model id
TEST_MODEL_ID = "c8a5b29a-cdba-408a-803a-a605a95631c2"


@pytest.fixture()
def random_str(rand_gen):
    return rand_gen(str)


@pytest.fixture(scope="module")
def foundry_client(client):
    return FoundryClient(client)


@pytest.fixture()
def text_data_row(dataset, random_str):
    global_key = "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/sample-text-1.txt-{random_str}"
    task = dataset.create_data_rows(
        [
            {
                "row_data": "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/sample-text-1.txt",
                "media_type": "TEXT",
                "global_key": global_key,
            }
        ]
    )
    task.wait_till_done()
    dr = dataset.data_rows().get_one()
    yield dr
    dr.delete()


@pytest.fixture()
def ontology(client, random_str):
    object_features = [
        lb.Tool(
            tool=lb.Tool.Type.BBOX,
            name="text",
            color="#ff0000",
            classifications=[
                lb.Classification(
                    class_type=lb.Classification.Type.TEXT, name="value"
                )
            ],
        )
    ]

    ontology_builder = lb.OntologyBuilder(
        tools=object_features,
    )

    ontology = client.create_ontology(
        f"Test ontology for tesseract model {random_str}",
        ontology_builder.asdict(),
        media_type=lb.MediaType.Image,
    )
    return ontology


@pytest.fixture()
def unsaved_app(random_str, ontology):
    return App(
        model_id=TEST_MODEL_ID,
        name=f"Test App {random_str}",
        description="Test App Description",
        inference_params={"confidence": 0.2},
        class_to_schema_id={},
        ontology_id=ontology.uid,
    )


@pytest.fixture()
def app(foundry_client, unsaved_app):
    app = foundry_client._create_app(unsaved_app)
    yield app
    foundry_client._delete_app(app.id)


def test_create_app(foundry_client, unsaved_app):
    app = foundry_client._create_app(unsaved_app)
    retrieved_dict = app.model_dump(exclude={"id", "created_by"})
    expected_dict = app.model_dump(exclude={"id", "created_by"})
    assert retrieved_dict == expected_dict


def test_get_app(foundry_client, app):
    retrieved_app = foundry_client._get_app(app.id)
    retrieved_dict = retrieved_app.model_dump(exclude={"created_by"})
    expected_dict = app.model_dump(exclude={"created_by"})
    assert retrieved_dict == expected_dict


def test_get_app_with_invalid_id(foundry_client):
    with pytest.raises(ResourceNotFoundError):
        foundry_client._get_app("invalid-id")


def test_run_foundry_app_with_data_row_id(
    foundry_client, data_row, app, random_str
):
    data_rows = lb.DataRowIds([data_row.uid])
    task = foundry_client.run_app(
        model_run_name=f"test-app-with-datarow-id-{random_str}",
        data_rows=data_rows,
        app_id=app.id,
    )
    task.wait_till_done()
    assert task.status == "COMPLETE"


def test_run_foundry_app_with_global_key(
    foundry_client, data_row, app, random_str
):
    data_rows = lb.GlobalKeys([data_row.global_key])
    task = foundry_client.run_app(
        model_run_name=f"test-app-with-global-key-{random_str}",
        data_rows=data_rows,
        app_id=app.id,
    )
    task.wait_till_done()
    assert task.status == "COMPLETE"


def test_run_foundry_app_returns_model_run_id(
    foundry_client, data_row, app, random_str
):
    data_rows = lb.GlobalKeys([data_row.global_key])
    task = foundry_client.run_app(
        model_run_name=f"test-app-with-global-key-{random_str}",
        data_rows=data_rows,
        app_id=app.id,
    )
    model_run_id = task.metadata["modelRunId"]
    model_run = foundry_client.client.get_model_run(model_run_id)
    assert model_run.uid == model_run_id


def test_run_foundry_with_invalid_data_row_id(foundry_client, app, random_str):
    invalid_datarow_id = "invalid-global-key"
    data_rows = lb.GlobalKeys([invalid_datarow_id])
    with pytest.raises(LabelboxError) as exception:
        foundry_client.run_app(
            model_run_name=f"test-app-with-invalid-datarow-id-{random_str}",
            data_rows=data_rows,
            app_id=app.id,
        )
        assert invalid_datarow_id in exception.value


def test_run_foundry_with_invalid_global_key(foundry_client, app, random_str):
    invalid_global_key = "invalid-global-key"
    data_rows = lb.GlobalKeys([invalid_global_key])
    with pytest.raises(LabelboxError) as exception:
        foundry_client.run_app(
            model_run_name=f"test-app-with-invalid-global-key-{random_str}",
            data_rows=data_rows,
            app_id=app.id,
        )
        assert invalid_global_key in exception.value



from labelbox.data.annotation_types.annotation import ClassificationAnnotation
from labelbox.data.annotation_types.classification.classification import (
    Checklist,
    ClassificationAnswer,
    Radio,
    Text,
)
from labelbox.data.annotation_types.data import GenericDataRowData
from labelbox.data.annotation_types.label import Label

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_serialization():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=GenericDataRowData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="free_text_annotation",
                value=Text(confidence=0.5, answer="text_answer"),
            )
        ],
    )

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    assert res["confidence"] == 0.5
    assert res["name"] == "free_text_annotation"
    assert res["answer"] == "text_answer"
    assert res["dataRow"]["id"] == "bkj7z2q0b0000jx6x0q2q7q0d"


def test_nested_serialization():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=GenericDataRowData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
            text="This is a test",
        ),
        annotations=[
            ClassificationAnnotation(
                name="nested test",
                value=Checklist(
                    answer=[
                        ClassificationAnswer(
                            name="first_answer",
                            confidence=0.9,
                            classifications=[
                                ClassificationAnnotation(
                                    name="sub_radio_question",
                                    value=Radio(
                                        answer=ClassificationAnswer(
                                            name="first_sub_radio_answer",
                                            confidence=0.8,
                                            classifications=[
                                                ClassificationAnnotation(
                                                    name="nested answer",
                                                    value=Text(
                                                        answer="nested answer",
                                                        confidence=0.7,
                                                    ),
                                                )
                                            ],
                                        )
                                    ),
                                )
                            ],
                        )
                    ]
                ),
            )
        ],
    )

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    assert res["dataRow"]["id"] == "bkj7z2q0b0000jx6x0q2q7q0d"
    answer = res["answer"][0]
    assert answer["confidence"] == 0.9
    assert answer["name"] == "first_answer"
    classification = answer["classifications"][0]
    nested_classification_answer = classification["answer"]
    assert nested_classification_answer["confidence"] == 0.8
    assert nested_classification_answer["name"] == "first_sub_radio_answer"
    sub_classification = nested_classification_answer["classifications"][0]
    assert sub_classification["name"] == "nested answer"
    assert sub_classification["answer"] == "nested answer"
    assert sub_classification["confidence"] == 0.7



from labelbox.client import Client


# @patch.dict(os.environ, {'LABELBOX_API_KEY': 'bar'})
def test_headers():
    client = Client(api_key="api_key", endpoint="http://localhost:8080/_gql")
    assert client.headers
    assert client.headers["Authorization"] == "Bearer api_key"
    assert client.headers["Content-Type"] == "application/json"
    assert client.headers["User-Agent"]
    assert client.headers["X-Python-Version"]


def test_enable_experimental():
    client = Client(api_key="api_key", enable_experimental=True)
    assert client.enable_experimental



from enum import Enum
from typing import Optional, Tuple, Dict, Union

from pydantic import conint

from .base import ConfidenceValue, BaseMetric

Count = conint(ge=0, le=1e10)

ConfusionMatrixMetricValue = Tuple[Count, Count, Count, Count]
ConfusionMatrixMetricConfidenceValue = Dict[
    ConfidenceValue, ConfusionMatrixMetricValue
]


class ConfusionMatrixAggregation(Enum):
    CONFUSION_MATRIX = "CONFUSION_MATRIX"


class ConfusionMatrixMetric(BaseMetric):
    """Class representing confusion matrix metrics.

    In the editor, this provides precision, recall, and f-scores.
    This should be used over multiple scalar metrics so that aggregations are accurate.

    Value should be a tuple representing:
      [True Positive Count, False Positive Count, True Negative Count, False Negative Count]

    aggregation cannot be adjusted for confusion matrix metrics.
    """

    metric_name: str
    value: Union[
        ConfusionMatrixMetricValue, ConfusionMatrixMetricConfidenceValue
    ]
    aggregation: Optional[ConfusionMatrixAggregation] = (
        ConfusionMatrixAggregation.CONFUSION_MATRIX
    )



import pytest

from labelbox.orm.model import Relationship
from labelbox.orm.db_object import DbObject


def test_illegal_cache_cond1():
    class TestEntityA(DbObject):
        test_entity_b = Relationship.ToOne("TestEntityB", cache=True)

    with pytest.raises(TypeError) as exc_info:

        class TestEntityB(DbObject):
            another_entity = Relationship.ToOne("AnotherEntity", cache=True)

    assert (
        "`test_entity_a` caches `test_entity_b` which caches `['another_entity']`"
        in str(exc_info.value)
    )


def test_illegal_cache_cond2():
    class TestEntityD(DbObject):
        another_entity = Relationship.ToOne("AnotherEntity", cache=True)

    with pytest.raises(TypeError) as exc_info:

        class TestEntityC(DbObject):
            test_entity_d = Relationship.ToOne("TestEntityD", cache=True)

    assert (
        "`test_entity_c` caches `test_entity_d` which caches `['another_entity']`"
        in str(exc_info.value)
    )



from labelbox.data.annotation_types.annotation import ClassificationAnnotation
from labelbox.data.annotation_types.classification.classification import (
    Checklist,
    ClassificationAnswer,
    Radio,
)
from labelbox.data.annotation_types.data import GenericDataRowData
from labelbox.data.annotation_types.label import Label

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_serialization_min():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=GenericDataRowData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
        ),
        annotations=[
            ClassificationAnnotation(
                name="checkbox_question_geo",
                value=Checklist(
                    answer=[ClassificationAnswer(name="first_answer")]
                ),
            )
        ],
    )

    expected = {
        "name": "checkbox_question_geo",
        "dataRow": {"id": "bkj7z2q0b0000jx6x0q2q7q0d"},
        "answer": [{"name": "first_answer"}],
    }
    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)
    res.pop("uuid")
    assert res == expected


def test_serialization_with_classification():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=GenericDataRowData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
        ),
        annotations=[
            ClassificationAnnotation(
                name="checkbox_question_geo",
                confidence=0.5,
                value=Checklist(
                    answer=[
                        ClassificationAnswer(
                            name="first_answer",
                            confidence=0.1,
                            classifications=[
                                ClassificationAnnotation(
                                    name="sub_radio_question",
                                    value=Radio(
                                        answer=ClassificationAnswer(
                                            name="first_sub_radio_answer",
                                            confidence=0.31,
                                        )
                                    ),
                                ),
                                ClassificationAnnotation(
                                    name="sub_chck_question",
                                    value=Checklist(
                                        answer=[
                                            ClassificationAnswer(
                                                name="second_subchk_answer",
                                                confidence=0.41,
                                            ),
                                            ClassificationAnswer(
                                                name="third_subchk_answer",
                                                confidence=0.42,
                                            ),
                                        ],
                                    ),
                                ),
                            ],
                        ),
                    ]
                ),
            )
        ],
    )

    expected = {
        "confidence": 0.5,
        "name": "checkbox_question_geo",
        "dataRow": {"id": "bkj7z2q0b0000jx6x0q2q7q0d"},
        "answer": [
            {
                "confidence": 0.1,
                "name": "first_answer",
                "classifications": [
                    {
                        "name": "sub_radio_question",
                        "answer": {
                            "confidence": 0.31,
                            "name": "first_sub_radio_answer",
                        },
                    },
                    {
                        "name": "sub_chck_question",
                        "answer": [
                            {
                                "confidence": 0.41,
                                "name": "second_subchk_answer",
                            },
                            {
                                "confidence": 0.42,
                                "name": "third_subchk_answer",
                            },
                        ],
                    },
                ],
            }
        ],
    }

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    res.pop("uuid")
    assert res == expected


def test_serialization_with_classification_double_nested():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=GenericDataRowData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
        ),
        annotations=[
            ClassificationAnnotation(
                name="checkbox_question_geo",
                confidence=0.5,
                value=Checklist(
                    answer=[
                        ClassificationAnswer(
                            name="first_answer",
                            confidence=0.1,
                            classifications=[
                                ClassificationAnnotation(
                                    name="sub_radio_question",
                                    value=Radio(
                                        answer=ClassificationAnswer(
                                            name="first_sub_radio_answer",
                                            confidence=0.31,
                                            classifications=[
                                                ClassificationAnnotation(
                                                    name="sub_chck_question",
                                                    value=Checklist(
                                                        answer=[
                                                            ClassificationAnswer(
                                                                name="second_subchk_answer",
                                                                confidence=0.41,
                                                            ),
                                                            ClassificationAnswer(
                                                                name="third_subchk_answer",
                                                                confidence=0.42,
                                                            ),
                                                        ],
                                                    ),
                                                )
                                            ],
                                        )
                                    ),
                                ),
                            ],
                        ),
                    ]
                ),
            )
        ],
    )

    expected = {
        "confidence": 0.5,
        "name": "checkbox_question_geo",
        "dataRow": {"id": "bkj7z2q0b0000jx6x0q2q7q0d"},
        "answer": [
            {
                "confidence": 0.1,
                "name": "first_answer",
                "classifications": [
                    {
                        "name": "sub_radio_question",
                        "answer": {
                            "confidence": 0.31,
                            "name": "first_sub_radio_answer",
                            "classifications": [
                                {
                                    "name": "sub_chck_question",
                                    "answer": [
                                        {
                                            "confidence": 0.41,
                                            "name": "second_subchk_answer",
                                        },
                                        {
                                            "confidence": 0.42,
                                            "name": "third_subchk_answer",
                                        },
                                    ],
                                }
                            ],
                        },
                    }
                ],
            }
        ],
    }
    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)

    res.pop("uuid")
    assert res == expected


def test_serialization_with_classification_double_nested_2():
    label = Label(
        uid="ckj7z2q0b0000jx6x0q2q7q0d",
        data=GenericDataRowData(
            uid="bkj7z2q0b0000jx6x0q2q7q0d",
        ),
        annotations=[
            ClassificationAnnotation(
                name="sub_radio_question",
                value=Radio(
                    answer=ClassificationAnswer(
                        name="first_sub_radio_answer",
                        confidence=0.31,
                        classifications=[
                            ClassificationAnnotation(
                                name="sub_chck_question",
                                value=Checklist(
                                    answer=[
                                        ClassificationAnswer(
                                            name="second_subchk_answer",
                                            confidence=0.41,
                                            classifications=[
                                                ClassificationAnnotation(
                                                    name="checkbox_question_geo",
                                                    value=Checklist(
                                                        answer=[
                                                            ClassificationAnswer(
                                                                name="first_answer",
                                                                confidence=0.1,
                                                            ),
                                                        ]
                                                    ),
                                                )
                                            ],
                                        ),
                                        ClassificationAnswer(
                                            name="third_subchk_answer",
                                            confidence=0.42,
                                        ),
                                    ]
                                ),
                            )
                        ],
                    )
                ),
            ),
        ],
    )

    expected = {
        "name": "sub_radio_question",
        "answer": {
            "confidence": 0.31,
            "name": "first_sub_radio_answer",
            "classifications": [
                {
                    "name": "sub_chck_question",
                    "answer": [
                        {
                            "confidence": 0.41,
                            "name": "second_subchk_answer",
                            "classifications": [
                                {
                                    "name": "checkbox_question_geo",
                                    "answer": [
                                        {
                                            "confidence": 0.1,
                                            "name": "first_answer",
                                        }
                                    ],
                                }
                            ],
                        },
                        {
                            "confidence": 0.42,
                            "name": "third_subchk_answer",
                        },
                    ],
                }
            ],
        },
        "dataRow": {"id": "bkj7z2q0b0000jx6x0q2q7q0d"},
    }

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)
    res.pop("uuid")
    assert res == expected



from dataclasses import dataclass, field
from enum import Enum

from labelbox.schema.tool_building.base_step_reasoning_tool import (
    _BaseStepReasoningTool,
    _Definition,
    _Variant,
)
from labelbox.schema.tool_building.tool_type import ToolType


class IncorrectStepActions(Enum):
    REGENERATE_STEPS = "regenerateSteps"
    GENERATE_AND_RATE_ALTERNATIVE_STEPS = "generateAndRateAlternativeSteps"
    REWRITE_STEP = "rewriteStep"
    JUSTIFICATION = "justification"


def build_step_reasoning_definition():
    correct_step = _Variant(id=0, name="Correct", actions=[])
    neutral_step = _Variant(id=1, name="Neutral", actions=[])
    incorrect_step = _Variant(
        id=2,
        name="Incorrect",
        _available_actions={action.value for action in IncorrectStepActions},
        actions=[action.value for action in IncorrectStepActions],
    )
    variants = [correct_step, neutral_step, incorrect_step]
    return _Definition(variants=variants)


@dataclass
class StepReasoningTool(_BaseStepReasoningTool):
    """
    Use this class in OntologyBuilder to create a tool for step reasoning
    The definition field lists the possible options to evaulate a step

    NOTE: color attribute is for backward compatibility only and should not be set directly
    """

    type: ToolType = field(default=ToolType.STEP_REASONING, init=False)
    definition: _Definition = field(
        default_factory=build_step_reasoning_definition
    )

    def __post_init__(self):
        super().__post_init__()
        # Set available actions for variants 0, 1, 2 'out of band' since they are not passed in the definition
        self._set_variant_available_actions()

    def _set_variant_available_actions(self):
        for variant in self.definition.variants:
            if variant.id == 2:
                for action in IncorrectStepActions:
                    variant._available_actions.add(action.value)



import random

from labelbox import StreamType


class TestExportEmbeddings:
    def test_export_embeddings_precomputed(
        self, client, dataset, environ, image_url
    ):
        data_row_specs = [
            {
                "row_data": image_url,
                "external_id": "image",
            }
        ]
        task = dataset.create_data_rows(data_row_specs)
        task.wait_till_done()
        export_task = dataset.export(params={"embeddings": True})
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert export_task.has_result()
        assert export_task.has_errors() is False

        results = []
        export_task.get_buffered_stream(stream_type=StreamType.RESULT).start(
            stream_handler=lambda output: results.append(output.json)
        )

        assert len(results) == len(data_row_specs)

        result = results[0]
        assert "embeddings" in result
        assert len(result["embeddings"]) > 0
        assert (
            result["embeddings"][0]["name"]
            == "Image Embedding V2 (CLIP ViT-B/32)"
        )
        assert len(result["embeddings"][0]["values"]) == 1

    def test_export_embeddings_custom(
        self, client, dataset, image_url, embedding
    ):
        vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
        import_task = dataset.create_data_rows(
            [
                {
                    "row_data": image_url,
                    "embeddings": [
                        {
                            "embedding_id": embedding.id,
                            "vector": vector,
                        }
                    ],
                }
            ]
        )
        import_task.wait_till_done()
        assert import_task.status == "COMPLETE"

        export_task = dataset.export(params={"embeddings": True})
        export_task.wait_till_done()
        assert export_task.status == "COMPLETE"
        assert export_task.has_result()
        assert export_task.has_errors() is False

        results = []
        export_task.get_buffered_stream(stream_type=StreamType.RESULT).start(
            stream_handler=lambda output: results.append(output.json)
        )

        assert len(results) == 1
        assert "embeddings" in results[0]
        assert (
            len(results[0]["embeddings"]) >= 1
        )  # should at least contain the custom embedding
        for emb in results[0]["embeddings"]:
            if emb["id"] == embedding.id:
                assert emb["name"] == embedding.name
                assert emb["dimensions"] == embedding.dims
                assert emb["is_custom"] is True
                assert len(emb["values"]) == 1
                assert emb["values"][0]["value"] == vector



from typing import Any, Dict, List, Union, Optional
from labelbox.data.annotation_types.base_annotation import BaseAnnotation

from labelbox.data.mixins import ConfidenceMixin, CustomMetricsMixin

from pydantic import BaseModel
from ..feature import FeatureSchema


class ClassificationAnswer(FeatureSchema, ConfidenceMixin, CustomMetricsMixin):
    """
    - Represents a classification option.
    - Because it inherits from FeatureSchema
        the option can be represented with either the name or feature_schema_id

    - The keyframe arg only applies to video classifications.
      Each answer can have a keyframe independent of the others.
        So unlike object annotations, classification annotations
          track keyframes at a classification answer level.
    """

    extra: Dict[str, Any] = {}
    keyframe: Optional[bool] = None
    classifications: Optional[List["ClassificationAnnotation"]] = None


class Radio(ConfidenceMixin, CustomMetricsMixin, BaseModel):
    """A classification with only one selected option allowed

    >>> Radio(answer = ClassificationAnswer(name = "dog"))

    """

    answer: ClassificationAnswer


class Checklist(ConfidenceMixin, BaseModel):
    """A classification with many selected options allowed

    >>> Checklist(answer = [ClassificationAnswer(name = "cloudy")])

    """

    answer: List[ClassificationAnswer]


class Text(ConfidenceMixin, CustomMetricsMixin, BaseModel):
    """Free form text

    >>> Text(answer = "some text answer")

    """

    answer: str


class ClassificationAnnotation(
    BaseAnnotation, ConfidenceMixin, CustomMetricsMixin
):
    """Classification annotations (non localized)

    >>> ClassificationAnnotation(
    >>>     value=Text(answer="my caption message"),
    >>>     feature_schema_id="my-feature-schema-id"
    >>> )

    Args:
        name (Optional[str])
        classifications (Optional[List[ClassificationAnnotation]]): Optional sub classification of the annotation
        feature_schema_id (Optional[Cuid])
        value (Union[Text, Checklist, Radio])
        extra (Dict[str, Any])
    """

    value: Union[Text, Checklist, Radio]
    message_id: Optional[str] = None



from typing import Any, Dict, List, Set


class LabelsConfidencePresenceChecker:
    """
    Checks if a given list of labels contains at least one confidence score
    """

    @classmethod
    def check(cls, raw_labels: List[Dict[str, Any]]):
        keys: Set[str] = set([])
        cls._collect_keys_from_list(raw_labels, keys)
        return len(keys.intersection(set(["confidence"]))) == 1

    @classmethod
    def _collect_keys_from_list(
        cls, objects: List[Dict[str, Any]], keys: Set[str]
    ):
        for obj in objects:
            if isinstance(obj, (list, tuple)):
                cls._collect_keys_from_list(obj, keys)
            elif isinstance(obj, dict):
                cls._collect_keys_from_object(obj, keys)

    @classmethod
    def _collect_keys_from_object(cls, object: Dict[str, Any], keys: Set[str]):
        for key in object:
            keys.add(key)
            if isinstance(object[key], dict):
                cls._collect_keys_from_object(object[key], keys)
            if isinstance(object[key], (list, tuple)):
                cls._collect_keys_from_list(object[key], keys)



from typing import List, Optional, Tuple

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    field_validator,
    model_validator,
)

from labelbox.data.annotation_types.annotation import (
    ClassificationAnnotation,
    ObjectAnnotation,
)
from labelbox.data.annotation_types.feature import FeatureSchema
from labelbox.data.mixins import (
    ConfidenceNotSupportedMixin,
    CustomMetricsNotSupportedMixin,
)
from labelbox.utils import _CamelCaseMixin, is_valid_uri


class VideoClassificationAnnotation(ClassificationAnnotation):
    """Video classification
    Args:
        name (Optional[str])
        feature_schema_id (Optional[Cuid])
        value (Union[Text, Checklist, Radio])
        frame (int): The frame index that this annotation corresponds to
        segment_id (Optional[Int]): Index of video segment this annotation belongs to
        extra (Dict[str, Any])
    """

    frame: int
    segment_index: Optional[int] = None


class VideoObjectAnnotation(
    ObjectAnnotation,
    ConfidenceNotSupportedMixin,
    CustomMetricsNotSupportedMixin,
):
    """Video object annotation
    >>> VideoObjectAnnotation(
    >>>     keyframe=True,
    >>>     frame=10,
    >>>     value=Rectangle(
    >>>        start=Point(x=0, y=0),
    >>>        end=Point(x=1, y=1)
    >>>     ),
    >>>     feature_schema_id="my-feature-schema-id"
    >>> )
    Args:
        name (Optional[str])
        feature_schema_id (Optional[Cuid])
        value (Geometry)
        frame (Int): The frame index that this annotation corresponds to
        keyframe (bool): Whether or not this annotation was a human generated or interpolated annotation
        segment_id (Optional[Int]): Index of video segment this annotation belongs to
        classifications (List[ClassificationAnnotation]) = []
        extra (Dict[str, Any])
    """

    frame: int
    keyframe: bool
    segment_index: Optional[int] = None


class MaskFrame(_CamelCaseMixin, BaseModel):
    index: int
    instance_uri: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("instanceURI", "instanceUri"),
        serialization_alias="instanceURI",
    )
    im_bytes: Optional[bytes] = None

    model_config = ConfigDict(populate_by_name=True)

    @model_validator(mode="after")
    def validate_args(self, values):
        im_bytes = self.im_bytes
        instance_uri = self.instance_uri
        if im_bytes == instance_uri is None:
            raise ValueError("One of `instance_uri`, `im_bytes` required.")
        return self

    @field_validator("instance_uri")
    def validate_uri(cls, v):
        if not is_valid_uri(v):
            raise ValueError(f"{v} is not a valid uri")
        return v


class MaskInstance(_CamelCaseMixin, FeatureSchema):
    color_rgb: Tuple[int, int, int] = Field(
        validation_alias=AliasChoices("colorRGB", "colorRgb"),
        serialization_alias="colorRGB",
    )
    name: str

    model_config = ConfigDict(populate_by_name=True)


class VideoMaskAnnotation(BaseModel):
    """Video mask annotation
    >>> VideoMaskAnnotation(
    >>>     frames=[
    >>>         MaskFrame(index=1, instance_uri='https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'),
    >>>         MaskFrame(index=5, instance_uri='https://storage.labelbox.com/cjhfn5y6s0pk507024nz1ocys1%2F1d60856c-59b7-3060-2754-83f7e93e0d01-1?Expires=1666901963361&KeyName=labelbox-assets-key-3&Signature=t-2s2DB4YjFuWEFak0wxYqfBfZA'),
    >>>     ],
    >>>     instances=[
    >>>         MaskInstance(color_rgb=(0, 0, 255), name="mask1"),
    >>>         MaskInstance(color_rgb=(0, 255, 0), name="mask2"),
    >>>         MaskInstance(color_rgb=(255, 0, 0), name="mask3")
    >>>     ]
    >>> )
    """

    frames: List[MaskFrame]
    instances: List[MaskInstance]



name = "labelbox"

__version__ = "6.5.0"

from labelbox.client import Client
from labelbox.schema.annotation_import import (
    LabelImport,
    MALPredictionImport,
    MEAPredictionImport,
    MEAToMALPredictionImport,
)
from labelbox.schema.asset_attachment import AssetAttachment
from labelbox.schema.batch import Batch
from labelbox.schema.benchmark import Benchmark
from labelbox.schema.catalog import Catalog
from labelbox.schema.data_row import DataRow
from labelbox.schema.data_row_metadata import (
    DataRowMetadata,
    DataRowMetadataField,
    DataRowMetadataOntology,
    DeleteDataRowMetadata,
)
from labelbox.schema.dataset import Dataset
from labelbox.schema.enums import AnnotationImportState
from labelbox.schema.export_task import (
    BufferedJsonConverterOutput,
    ExportTask,
    BufferedJsonConverterOutput,
    StreamType,
)
from labelbox.schema.iam_integration import IAMIntegration
from labelbox.schema.identifiable import GlobalKey, UniqueId
from labelbox.schema.identifiables import DataRowIds, GlobalKeys, UniqueIds
from labelbox.schema.invite import Invite, InviteLimit
from labelbox.schema.label import Label
from labelbox.schema.label_score import LabelScore
from labelbox.schema.labeling_frontend import (
    LabelingFrontend,
    LabelingFrontendOptions,
)
from labelbox.schema.labeling_service import LabelingService
from labelbox.schema.labeling_service_dashboard import LabelingServiceDashboard
from labelbox.schema.labeling_service_status import LabelingServiceStatus
from labelbox.schema.media_type import MediaType
from labelbox.schema.model import Model
from labelbox.schema.model_config import ModelConfig
from labelbox.schema.model_run import DataSplit, ModelRun
from labelbox.schema.ontology import (
    FeatureSchema,
    Ontology,
    OntologyBuilder,
    Tool,
)
from labelbox.schema.tool_building.fact_checking_tool import FactCheckingTool
from labelbox.schema.tool_building.step_reasoning_tool import StepReasoningTool
from labelbox.schema.tool_building.prompt_issue_tool import PromptIssueTool
from labelbox.schema.role import Role, ProjectRole
from labelbox.schema.invite import Invite, InviteLimit
from labelbox.schema.data_row_metadata import (
    DataRowMetadataOntology,
    DataRowMetadataField,
    DataRowMetadata,
    DeleteDataRowMetadata,
)
from labelbox.schema.model_run import ModelRun, DataSplit
from labelbox.schema.benchmark import Benchmark
from labelbox.schema.iam_integration import IAMIntegration
from labelbox.schema.resource_tag import ResourceTag
from labelbox.schema.project_model_config import ProjectModelConfig
from labelbox.schema.project_resource_tag import ProjectResourceTag
from labelbox.schema.media_type import MediaType
from labelbox.schema.slice import Slice, CatalogSlice, ModelSlice
from labelbox.schema.task_queue import TaskQueue
from labelbox.schema.label_score import LabelScore
from labelbox.schema.identifiables import UniqueIds, GlobalKeys, DataRowIds
from labelbox.schema.identifiable import UniqueId, GlobalKey
from labelbox.schema.ontology_kind import OntologyKind
from labelbox.schema.organization import Organization
from labelbox.schema.project import Project
from labelbox.schema.project_model_config import ProjectModelConfig
from labelbox.schema.project_overview import (
    ProjectOverview,
    ProjectOverviewDetailed,
)
from labelbox.schema.project_resource_tag import ProjectResourceTag
from labelbox.schema.resource_tag import ResourceTag
from labelbox.schema.review import Review
from labelbox.schema.role import ProjectRole, Role
from labelbox.schema.slice import CatalogSlice, ModelSlice, Slice
from labelbox.schema.task import Task
from labelbox.schema.task_queue import TaskQueue
from labelbox.schema.user import User
from labelbox.schema.webhook import Webhook
from labelbox.schema.tool_building.classification import (
    Classification,
    Option,
    ResponseOption,
    PromptResponseClassification,
)
from lbox.exceptions import *



from dataclasses import dataclass
from typing import Optional, Tuple, Union
import warnings
from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field
from labelbox.pagination import PaginatedCollection
from labelbox.schema.export_params import (
    CatalogExportParams,
    validate_catalog_export_params,
)
from labelbox.schema.export_task import ExportTask
from labelbox.schema.identifiable import GlobalKey, UniqueId
from labelbox.schema.task import Task


class Slice(DbObject):
    """
    A Slice is a saved set of filters (saved query).
    This is an abstract class and should not be instantiated.

    Attributes:
        name (datetime)
        description (datetime)
        created_at (datetime)
        updated_at (datetime)
        filter (json)
    """

    name = Field.String("name")
    description = Field.String("description")
    created_at = Field.DateTime("created_at")
    updated_at = Field.DateTime("updated_at")
    filter = Field.Json("filter")

    @dataclass
    class DataRowIdAndGlobalKey:
        id: UniqueId
        global_key: Optional[GlobalKey]

        def __init__(self, id: str, global_key: Optional[str]):
            self.id = UniqueId(id)
            self.global_key = GlobalKey(global_key) if global_key else None

        def to_hash(self):
            return {
                "id": self.id.key,
                "global_key": self.global_key.key if self.global_key else None,
            }


class CatalogSlice(Slice):
    """
    Represents a Slice used for filtering data rows in Catalog.
    """

    def get_data_row_identifiers(self) -> PaginatedCollection:
        """
        Fetches all data row ids and global keys (where defined) that match this Slice

        Returns:
            A PaginatedCollection of Slice.DataRowIdAndGlobalKey
        """
        query_str = """
            query getDataRowIdenfifiersBySavedQueryPyApi($id: ID!, $from: String, $first: Int!) {
                getDataRowIdentifiersBySavedQuery(input: {
                    savedQueryId: $id,
                    after: $from
                    first: $first
                }) {
                    totalCount
                    nodes
                    {
                        id
                        globalKey
                    }
                    pageInfo {
                        endCursor
                        hasNextPage
                    }
                }
            }
        """
        return PaginatedCollection(
            client=self.client,
            query=query_str,
            params={"id": str(self.uid)},
            dereferencing=["getDataRowIdentifiersBySavedQuery", "nodes"],
            obj_class=lambda _, data_row_id_and_gk: Slice.DataRowIdAndGlobalKey(
                data_row_id_and_gk.get("id"),
                data_row_id_and_gk.get("globalKey", None),
            ),
            cursor_path=[
                "getDataRowIdentifiersBySavedQuery",
                "pageInfo",
                "endCursor",
            ],
        )

    def export(
        self,
        task_name: Optional[str] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> ExportTask:
        """
        Creates a slice export task with the given params and returns the task.
        >>>     slice = client.get_catalog_slice("SLICE_ID")
        >>>     task = slice.export(
        >>>         params={"performance_details": False, "label_details": True}
        >>>     )
        >>>     task.wait_till_done()
        >>>     task.result
        """
        task, _ = self._export(task_name, params, streamable=True)
        return ExportTask(task)

    def export_v2(
        self,
        task_name: Optional[str] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> Union[Task, ExportTask]:
        """
        Creates a slice export task with the given params and returns the task.
        >>>     slice = client.get_catalog_slice("SLICE_ID")
        >>>     task = slice.export_v2(
        >>>         params={"performance_details": False, "label_details": True}
        >>>     )
        >>>     task.wait_till_done()
        >>>     task.result
        """

        warnings.warn(
            "The method export_v2 for CatalogSlice is deprecated and will be removed in the next major release. Use the export method instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        task, is_streamable = self._export(task_name, params)
        if is_streamable:
            return ExportTask(task, True)
        return task

    def _export(
        self,
        task_name: Optional[str] = None,
        params: Optional[CatalogExportParams] = None,
        streamable: bool = False,
    ) -> Tuple[Task, bool]:
        _params = params or CatalogExportParams(
            {
                "attachments": False,
                "embeddings": False,
                "metadata_fields": False,
                "data_row_details": False,
                "project_details": False,
                "performance_details": False,
                "label_details": False,
                "media_type_override": None,
                "model_run_ids": None,
                "project_ids": None,
                "interpolated_frames": False,
                "all_projects": False,
                "all_model_runs": False,
                "predictions": False,
            }
        )
        validate_catalog_export_params(_params)

        mutation_name = "exportDataRowsInSlice"
        create_task_query_str = (
            f"mutation {mutation_name}PyApi"
            f"($input: ExportDataRowsInSliceInput!)"
            f"{{{mutation_name}(input: $input){{taskId isStreamable}}}}"
        )

        media_type_override = _params.get("media_type_override", None)
        query_params = {
            "input": {
                "taskName": task_name,
                "filters": {"sliceId": self.uid},
                "isStreamableReady": True,
                "params": {
                    "mediaTypeOverride": media_type_override.value
                    if media_type_override is not None
                    else None,
                    "includeAttachments": _params.get("attachments", False),
                    "includeEmbeddings": _params.get("embeddings", False),
                    "includeMetadata": _params.get("metadata_fields", False),
                    "includeDataRowDetails": _params.get(
                        "data_row_details", False
                    ),
                    "includeProjectDetails": _params.get(
                        "project_details", False
                    ),
                    "includePerformanceDetails": _params.get(
                        "performance_details", False
                    ),
                    "includeLabelDetails": _params.get("label_details", False),
                    "includeInterpolatedFrames": _params.get(
                        "interpolated_frames", False
                    ),
                    "projectIds": _params.get("project_ids", None),
                    "modelRunIds": _params.get("model_run_ids", None),
                    "allProjects": _params.get("all_projects", False),
                    "allModelRuns": _params.get("all_model_runs", False),
                },
                "streamable": streamable,
            }
        }

        res = self.client.execute(
            create_task_query_str, query_params, error_log_key="errors"
        )
        res = res[mutation_name]
        task_id = res["taskId"]
        is_streamable = res["isStreamable"]
        return Task.get_task(self.client, task_id), is_streamable


class ModelSlice(Slice):
    """
    Represents a Slice used for filtering data rows in Model.
    """

    @classmethod
    def query_str(cls):
        query_str = """
        query getDataRowIdenfifiersBySavedModelQueryPyApi($id: ID!, $modelRunId: ID, $from: DataRowIdentifierCursorInput, $first: Int!) {
            getDataRowIdentifiersBySavedModelQuery(input: {
                savedQueryId: $id,
                modelRunId: $modelRunId,
                after: $from
                first: $first
            }) {
                totalCount
                nodes
                {
                    id
                    globalKey
                }
                pageInfo {
                    endCursor {
                        dataRowId
                        globalKey
                    }
                    hasNextPage
                }
            }
        }
        """
        return query_str

    def get_data_row_ids(self, model_run_id: str) -> PaginatedCollection:
        """
        Fetches all data row ids that match this Slice

        Params
        model_run_id: str, required, uid or cuid of model run

        Returns:
            A PaginatedCollection of data row ids
        """

        warnings.warn(
            "The method get_data_row_ids for ModelSlice is deprecated and will be removed in the next major release. Use the get_data_row_identifiers method instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return PaginatedCollection(
            client=self.client,
            query=ModelSlice.query_str(),
            params={"id": str(self.uid), "modelRunId": model_run_id},
            dereferencing=["getDataRowIdentifiersBySavedModelQuery", "nodes"],
            obj_class=lambda _, data_row_id_and_gk: data_row_id_and_gk.get(
                "id"
            ),
            cursor_path=[
                "getDataRowIdentifiersBySavedModelQuery",
                "pageInfo",
                "endCursor",
            ],
        )

    def get_data_row_identifiers(
        self, model_run_id: str
    ) -> PaginatedCollection:
        """
        Fetches all data row ids and global keys (where defined) that match this Slice

        Params:
        model_run_id : str, required, uid or cuid of model run

        Returns:
            A PaginatedCollection of Slice.DataRowIdAndGlobalKey
        """
        return PaginatedCollection(
            client=self.client,
            query=ModelSlice.query_str(),
            params={"id": str(self.uid), "modelRunId": model_run_id},
            dereferencing=["getDataRowIdentifiersBySavedModelQuery", "nodes"],
            obj_class=lambda _, data_row_id_and_gk: Slice.DataRowIdAndGlobalKey(
                data_row_id_and_gk.get("id"),
                data_row_id_and_gk.get("globalKey", None),
            ),
            cursor_path=[
                "getDataRowIdentifiersBySavedModelQuery",
                "pageInfo",
                "endCursor",
            ],
        )



import time
import uuid
from typing import Tuple, Type, Union

import pytest
import requests
from pytest import FixtureRequest

from labelbox import Client, Dataset, MediaType, OntologyKind, parser
from labelbox.schema.annotation_import import AnnotationImportState, LabelImport
from labelbox.schema.model_run import ModelRun
from labelbox.schema.ontology import Ontology
from labelbox.schema.project import Project

"""
The main fixtures of this library are configured_project and configured_project_by_global_key. Both fixtures generate data rows with a parametrize media type. They create the amount of data rows equal to the DATA_ROW_COUNT variable below. The data rows are generated with a factory fixture that returns a function that allows you to pass a global key. The ontologies are generated normalized and based on the MediaType given (i.e. only features supported by MediaType are created). This ontology is later used to obtain the correct annotations with the prediction_id_mapping and corresponding inferences. Each data row will have all possible annotations attached supported for the MediaType. 
"""

DATA_ROW_COUNT = 3
DATA_ROW_PROCESSING_WAIT_TIMEOUT_SECONDS = 40
DATA_ROW_PROCESSING_WAIT_SLEEP_INTERNAL_SECONDS = 7


@pytest.fixture(scope="module", autouse=True)
def video_data_row_factory():
    def video_data_row(global_key):
        return {
            "row_data": "https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4",
            "global_key": f"https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4-{global_key}",
            "media_type": "VIDEO",
        }

    return video_data_row


@pytest.fixture(scope="module", autouse=True)
def audio_data_row_factory():
    def audio_data_row(global_key):
        return {
            "row_data": "https://storage.googleapis.com/labelbox-datasets/audio-sample-data/sample-audio-1.mp3",
            "global_key": f"https://storage.googleapis.com/labelbox-datasets/audio-sample-data/sample-audio-1.mp3-{global_key}",
            "media_type": "AUDIO",
        }

    return audio_data_row


@pytest.fixture(scope="module", autouse=True)
def conversational_data_row_factory():
    def conversational_data_row(global_key):
        return {
            "row_data": "https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json",
            "global_key": f"https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json-{global_key}",
        }

    return conversational_data_row


@pytest.fixture(scope="module", autouse=True)
def geospatial_data_row_factory():
    def geospatial_data_row(global_key):
        return {
            "row_data": {
                "tile_layer_url": "https://s3-us-west-1.amazonaws.com/lb-tiler-layers/mexico_city/{z}/{x}/{y}.png",
                "bounds": [
                    [19.405662413477728, -99.21052827588443],
                    [19.400498983095076, -99.20534818927473],
                ],
                "min_zoom": 12,
                "max_zoom": 20,
                "epsg": "EPSG4326",
            },
            "global_key": f"https://s3-us-west-1.amazonaws.com/lb-tiler-layers/mexico_city/z/x/y.png-{global_key}",
            "media_type": "TMS_GEO",
        }

    return geospatial_data_row


@pytest.fixture(scope="module", autouse=True)
def html_data_row_factory():
    def html_data_row(global_key):
        return {
            "row_data": "https://storage.googleapis.com/labelbox-datasets/html_sample_data/sample_html_1.html",
            "global_key": f"https://storage.googleapis.com/labelbox-datasets/html_sample_data/sample_html_1.html-{global_key}",
        }

    return html_data_row


@pytest.fixture(scope="module", autouse=True)
def image_data_row_factory():
    def image_data_row(global_key):
        return {
            "row_data": "https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg",
            "global_key": f"https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-{global_key}",
            "media_type": "IMAGE",
        }

    return image_data_row


@pytest.fixture(scope="module", autouse=True)
def document_data_row_factory():
    def document_data_row(global_key):
        return {
            "row_data": {
                "pdf_url": "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf",
                "text_layer_url": "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483-lb-textlayer.json",
            },
            "global_key": f"https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf-{global_key}",
            "media_type": "PDF",
        }

    return document_data_row


@pytest.fixture(scope="module", autouse=True)
def text_data_row_factory():
    def text_data_row(global_key):
        return {
            "row_data": "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/sample-text-2.txt",
            "global_key": f"https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/sample-text-2.txt-{global_key}",
            "media_type": "TEXT",
        }

    return text_data_row


@pytest.fixture(scope="module", autouse=True)
def llm_human_preference_data_row_factory():
    def llm_human_preference_data_row(global_key):
        return {
            "row_data": "https://storage.googleapis.com/labelbox-datasets/sdk_test/llm_prompt_response_conv.json",
            "global_key": global_key,
        }

    return llm_human_preference_data_row


@pytest.fixture(scope="module")
def mmc_data_row_url():
    return "https://storage.googleapis.com/labelbox-datasets/conversational_model_evaluation_sample/offline-model-chat-evaluation.json"


@pytest.fixture(scope="module", autouse=True)
def offline_model_evaluation_data_row_factory(mmc_data_row_url: str):
    def offline_model_evaluation_data_row(global_key: str):
        return {
            "row_data": mmc_data_row_url,
            "global_key": global_key,
        }

    return offline_model_evaluation_data_row


@pytest.fixture(scope="module", autouse=True)
def data_row_json_by_media_type(
    audio_data_row_factory,
    conversational_data_row_factory,
    geospatial_data_row_factory,
    html_data_row_factory,
    image_data_row_factory,
    document_data_row_factory,
    text_data_row_factory,
    video_data_row_factory,
    offline_model_evaluation_data_row_factory,
):
    return {
        MediaType.Audio: audio_data_row_factory,
        MediaType.Conversational: conversational_data_row_factory,
        MediaType.Geospatial_Tile: geospatial_data_row_factory,
        MediaType.Html: html_data_row_factory,
        MediaType.Image: image_data_row_factory,
        MediaType.Document: document_data_row_factory,
        MediaType.Text: text_data_row_factory,
        MediaType.Video: video_data_row_factory,
        OntologyKind.ModelEvaluation: offline_model_evaluation_data_row_factory,
    }


@pytest.fixture(scope="module", autouse=True)
def normalized_ontology_by_media_type():
    """Returns NDJSON of ontology based on media type"""

    bbox_tool_with_nested_text = {
        "required": False,
        "name": "bbox_tool_with_nested_text",
        "tool": "rectangle",
        "color": "#a23030",
        "classifications": [
            {
                "required": False,
                "instructions": "nested",
                "name": "nested",
                "type": "radio",
                "options": [
                    {
                        "label": "radio_value_1",
                        "value": "radio_value_1",
                        "options": [
                            {
                                "required": False,
                                "instructions": "nested_checkbox",
                                "name": "nested_checkbox",
                                "type": "checklist",
                                "options": [
                                    {
                                        "label": "nested_checkbox_option_1",
                                        "value": "nested_checkbox_option_1",
                                        "options": [],
                                    },
                                    {
                                        "label": "nested_checkbox_option_2",
                                        "value": "nested_checkbox_option_2",
                                    },
                                ],
                            },
                            {
                                "required": False,
                                "instructions": "nested_text",
                                "name": "nested_text",
                                "type": "text",
                                "options": [],
                            },
                        ],
                    },
                ],
            }
        ],
    }

    bbox_tool = {
        "required": False,
        "name": "bbox",
        "tool": "rectangle",
        "color": "#a23030",
        "classifications": [],
    }

    polygon_tool = {
        "required": False,
        "name": "polygon",
        "tool": "polygon",
        "color": "#FF34FF",
        "classifications": [],
    }
    polyline_tool = {
        "required": False,
        "name": "polyline",
        "tool": "line",
        "color": "#FF4A46",
        "classifications": [],
    }
    point_tool = {
        "required": False,
        "name": "point--",
        "tool": "point",
        "color": "#008941",
        "classifications": [],
    }
    entity_tool = {
        "required": False,
        "name": "named-entity",
        "tool": "named-entity",
        "color": "#006FA6",
        "classifications": [],
    }
    raster_segmentation_tool = {
        "required": False,
        "name": "segmentation_mask",
        "tool": "raster-segmentation",
        "color": "#ff0000",
        "classifications": [],
    }
    segmentation_tool = {
        "required": False,
        "name": "segmentation--",
        "tool": "superpixel",
        "color": "#A30059",
        "classifications": [],
    }
    checklist = {
        "required": False,
        "instructions": "checklist",
        "name": "checklist",
        "type": "checklist",
        "options": [
            {
                "label": "first_checklist_answer",
                "value": "first_checklist_answer",
            },
            {
                "label": "second_checklist_answer",
                "value": "second_checklist_answer",
            },
        ],
    }
    checklist_index = {
        "required": False,
        "instructions": "checklist_index",
        "name": "checklist_index",
        "type": "checklist",
        "scope": "index",
        "options": [
            {
                "label": "first_checklist_answer",
                "value": "first_checklist_answer",
            },
            {
                "label": "second_checklist_answer",
                "value": "second_checklist_answer",
            },
        ],
    }
    free_form_text = {
        "required": False,
        "instructions": "text",
        "name": "text",
        "type": "text",
        "options": [],
    }
    free_form_text_index = {
        "required": False,
        "instructions": "text_index",
        "name": "text_index",
        "type": "text",
        "scope": "index",
        "options": [],
    }
    radio = {
        "required": False,
        "instructions": "radio",
        "name": "radio",
        "type": "radio",
        "options": [
            {
                "label": "first_radio_answer",
                "value": "first_radio_answer",
                "options": [],
            },
            {
                "label": "second_radio_answer",
                "value": "second_radio_answer",
                "options": [],
            },
        ],
    }

    radio_index = {
        "required": False,
        "instructions": "radio_index",
        "name": "radio_index",
        "type": "radio",
        "scope": "index",
        "options": [
            {
                "label": "first_radio_answer",
                "value": "first_radio_answer",
                "options": [],
            },
            {
                "label": "second_radio_answer",
                "value": "second_radio_answer",
                "options": [],
            },
        ],
    }

    prompt_text = {
        "instructions": "prompt-text",
        "name": "prompt-text",
        "options": [],
        "required": True,
        "maxCharacters": 50,
        "minCharacters": 1,
        "schemaNodeId": None,
        "type": "prompt",
    }

    response_radio = {
        "instructions": "radio-response",
        "name": "radio-response",
        "options": [
            {
                "label": "first_radio_answer",
                "value": "first_radio_answer",
                "options": [],
            },
            {
                "label": "second_radio_answer",
                "value": "second_radio_answer",
                "options": [],
            },
        ],
        "required": True,
        "type": "response-radio",
    }

    response_checklist = {
        "instructions": "checklist-response",
        "name": "checklist-response",
        "options": [
            {
                "label": "first_checklist_answer",
                "value": "first_checklist_answer",
                "options": [],
            },
            {
                "label": "second_checklist_answer",
                "value": "second_checklist_answer",
                "options": [],
            },
        ],
        "required": True,
        "type": "response-checklist",
    }

    response_text = {
        "instructions": "response-text",
        "maxCharacters": 20,
        "minCharacters": 1,
        "name": "response-text",
        "required": True,
        "type": "response-text",
    }

    message_single_selection_task = {
        "required": False,
        "name": "message-single-selection",
        "tool": "message-single-selection",
        "classifications": [],
    }

    message_multi_selection_task = {
        "required": False,
        "name": "message-multi-selection",
        "tool": "message-multi-selection",
        "classifications": [],
    }

    message_ranking_task = {
        "required": False,
        "name": "message-ranking",
        "tool": "message-ranking",
        "classifications": [],
    }

    return {
        MediaType.Image: {
            "tools": [
                bbox_tool,
                bbox_tool_with_nested_text,
                polygon_tool,
                polyline_tool,
                point_tool,
                raster_segmentation_tool,
            ],
            "classifications": [
                checklist,
                free_form_text,
                radio,
            ],
        },
        MediaType.Text: {
            "tools": [entity_tool],
            "classifications": [
                checklist,
                free_form_text,
                radio,
            ],
        },
        MediaType.Video: {
            "tools": [
                bbox_tool,
                bbox_tool_with_nested_text,
                polyline_tool,
                point_tool,
                raster_segmentation_tool,
            ],
            "classifications": [
                checklist,
                free_form_text,
                radio,
                checklist_index,
                free_form_text_index,
            ],
        },
        MediaType.Geospatial_Tile: {
            "tools": [
                bbox_tool,
                bbox_tool_with_nested_text,
                polygon_tool,
                polyline_tool,
                point_tool,
            ],
            "classifications": [
                checklist,
                free_form_text,
                radio,
            ],
        },
        MediaType.Document: {
            "tools": [entity_tool, bbox_tool, bbox_tool_with_nested_text],
            "classifications": [
                checklist,
                free_form_text,
                radio,
            ],
        },
        MediaType.Audio: {
            "tools": [],
            "classifications": [
                checklist,
                free_form_text,
                radio,
            ],
        },
        MediaType.Html: {
            "tools": [],
            "classifications": [
                checklist,
                free_form_text,
                radio,
            ],
        },
        MediaType.Conversational: {
            "tools": [entity_tool],
            "classifications": [
                checklist,
                free_form_text,
                radio,
                checklist_index,
                free_form_text_index,
            ],
        },
        MediaType.LLMPromptResponseCreation: {
            "tools": [],
            "classifications": [
                prompt_text,
                response_text,
                response_radio,
                response_checklist,
            ],
        },
        MediaType.LLMPromptCreation: {
            "tools": [],
            "classifications": [prompt_text],
        },
        OntologyKind.ResponseCreation: {
            "tools": [],
            "classifications": [
                response_text,
                response_radio,
                response_checklist,
            ],
        },
        OntologyKind.ModelEvaluation: {
            "tools": [
                message_single_selection_task,
                message_multi_selection_task,
                message_ranking_task,
            ],
            "classifications": [
                radio,
                checklist,
                free_form_text,
                radio_index,
                checklist_index,
                free_form_text_index,
            ],
        },
        "all": {
            "tools": [
                bbox_tool,
                bbox_tool_with_nested_text,
                polygon_tool,
                polyline_tool,
                point_tool,
                entity_tool,
                segmentation_tool,
                raster_segmentation_tool,
            ],
            "classifications": [
                checklist,
                checklist_index,
                free_form_text,
                free_form_text_index,
                radio,
            ],
        },
    }


@pytest.fixture
def wait_for_label_processing():
    """
    Do not use. Only for testing.

    Returns project's labels as a list after waiting for them to finish processing.
    If `project.labels()` is called before label is fully processed,
    it may return an empty set
    """

    def func(project):
        timeout_seconds = 10
        while True:
            labels = list(project.labels())
            if len(labels) > 0:
                return labels
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for label for project '{project.uid}' to finish processing"
                )
            time.sleep(2)

    return func


##### Unit test strategies #####


@pytest.fixture
def hardcoded_datarow_id():
    data_row_id = "ck8q9q9qj00003g5z3q1q9q9q"

    def get_data_row_id():
        return data_row_id

    yield get_data_row_id


@pytest.fixture
def hardcoded_global_key():
    global_key = str(uuid.uuid4())

    def get_global_key():
        return global_key

    yield get_global_key


##### Integration test strategies #####


def _create_response_creation_project(
    client: Client,
    rand_gen,
    data_row_json_by_media_type,
    ontology_kind,
    normalized_ontology_by_media_type,
) -> Tuple[Project, Ontology, Dataset]:
    "For response creation projects"

    dataset = client.create_dataset(name=rand_gen(str))

    project = client.create_response_creation_project(
        name=f"{ontology_kind}-{rand_gen(str)}"
    )

    ontology = client.create_ontology(
        name=f"{ontology_kind}-{rand_gen(str)}",
        normalized=normalized_ontology_by_media_type[ontology_kind],
        media_type=MediaType.Text,
        ontology_kind=ontology_kind,
    )

    project.connect_ontology(ontology)

    data_row_data = []

    for _ in range(DATA_ROW_COUNT):
        data_row_data.append(
            data_row_json_by_media_type[MediaType.Text](rand_gen(str))
        )

    task = dataset.create_data_rows(data_row_data)
    task.wait_till_done()
    global_keys = [row["global_key"] for row in task.result]
    data_row_ids = [row["id"] for row in task.result]

    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids
    project.global_keys = global_keys

    return project, ontology, dataset


@pytest.fixture
def llm_prompt_response_creation_dataset_with_data_row(
    client: Client, rand_gen
):
    dataset = client.create_dataset(name=rand_gen(str))
    global_key = str(uuid.uuid4())

    convo_data = {
        "row_data": "https://storage.googleapis.com/labelbox-datasets/conversational-sample-data/pairwise_shopping_2.json",
        "global_key": global_key,
    }

    task = dataset.create_data_rows([convo_data])
    task.wait_till_done()
    assert task.status == "COMPLETE"
    yield dataset

    dataset.delete()


def _create_prompt_response_project(
    client: Client,
    rand_gen,
    media_type,
    normalized_ontology_by_media_type,
    export_v2_test_helpers,
    llm_prompt_response_creation_dataset_with_data_row,
) -> Tuple[Project, Ontology]:
    """For prompt response data row auto gen projects"""
    dataset = llm_prompt_response_creation_dataset_with_data_row
    prompt_response_project = client.create_prompt_response_generation_project(
        name=f"{media_type.value}-{rand_gen(str)}",
        dataset_id=dataset.uid,
        media_type=media_type,
    )

    ontology = client.create_ontology(
        name=f"{media_type}-{rand_gen(str)}",
        normalized=normalized_ontology_by_media_type[media_type],
        media_type=media_type,
    )

    prompt_response_project.connect_ontology(ontology)

    # We have to export to get data row ids
    result = export_v2_test_helpers.run_project_export_v2_task(
        prompt_response_project
    )

    data_row_ids = [dr["data_row"]["id"] for dr in result]
    global_keys = [dr["data_row"]["global_key"] for dr in result]

    prompt_response_project.data_row_ids = data_row_ids
    prompt_response_project.global_keys = global_keys

    return prompt_response_project, ontology


def _create_offline_mmc_project(
    client: Client, rand_gen, data_row_json, normalized_ontology
) -> Tuple[Project, Ontology, Dataset]:
    dataset = client.create_dataset(name=rand_gen(str))

    project = client.create_offline_model_evaluation_project(
        name=f"offline-mmc-{rand_gen(str)}",
    )

    ontology = client.create_ontology(
        name=f"offline-mmc-{rand_gen(str)}",
        normalized=normalized_ontology,
        media_type=MediaType.Conversational,
        ontology_kind=OntologyKind.ModelEvaluation,
    )

    project.connect_ontology(ontology)

    data_row_data = [
        data_row_json(rand_gen(str)) for _ in range(DATA_ROW_COUNT)
    ]

    task = dataset.create_data_rows(data_row_data)
    task.wait_till_done()
    global_keys = [row["global_key"] for row in task.result]
    data_row_ids = [row["id"] for row in task.result]

    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids
    project.data_row_data = data_row_data
    project.global_keys = global_keys

    return project, ontology, dataset


def _create_project(
    client: Client,
    rand_gen,
    data_row_json_by_media_type,
    media_type,
    normalized_ontology_by_media_type,
) -> Tuple[Project, Ontology, Dataset]:
    """Shared function to configure project for integration tests"""

    dataset = client.create_dataset(name=rand_gen(str))

    project = client.create_project(
        name=f"{media_type}-{rand_gen(str)}", media_type=media_type
    )

    ontology = client.create_ontology(
        name=f"{media_type}-{rand_gen(str)}",
        normalized=normalized_ontology_by_media_type[media_type],
        media_type=media_type,
    )

    project.connect_ontology(ontology)
    data_row_data = []

    for _ in range(DATA_ROW_COUNT):
        data_row_data.append(
            data_row_json_by_media_type[media_type](rand_gen(str))
        )

    task = dataset.create_data_rows(data_row_data)
    task.wait_till_done()
    global_keys = [row["global_key"] for row in task.result]
    data_row_ids = [row["id"] for row in task.result]

    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids
    project.global_keys = global_keys

    return project, ontology, dataset


@pytest.fixture
def configured_project(
    client: Client,
    rand_gen,
    data_row_json_by_media_type,
    request: FixtureRequest,
    normalized_ontology_by_media_type,
    export_v2_test_helpers,
    llm_prompt_response_creation_dataset_with_data_row,
    teardown_helpers,
):
    """Configure project for test. Request.param will contain the media type if not present will use Image MediaType. The project will have 10 data rows."""

    media_type: Union[MediaType, OntologyKind] = getattr(
        request, "param", MediaType.Image
    )

    dataset = None

    if (
        media_type == MediaType.LLMPromptCreation
        or media_type == MediaType.LLMPromptResponseCreation
    ):
        project, ontology = _create_prompt_response_project(
            client,
            rand_gen,
            media_type,
            normalized_ontology_by_media_type,
            export_v2_test_helpers,
            llm_prompt_response_creation_dataset_with_data_row,
        )
    elif media_type == OntologyKind.ResponseCreation:
        project, ontology, dataset = _create_response_creation_project(
            client,
            rand_gen,
            data_row_json_by_media_type,
            media_type,
            normalized_ontology_by_media_type,
        )
    elif media_type == OntologyKind.ModelEvaluation:
        project, ontology, dataset = _create_offline_mmc_project(
            client,
            rand_gen,
            data_row_json_by_media_type[media_type],
            normalized_ontology_by_media_type[media_type],
        )
    else:
        project, ontology, dataset = _create_project(
            client,
            rand_gen,
            data_row_json_by_media_type,
            media_type,
            normalized_ontology_by_media_type,
        )

    yield project

    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)

    if dataset:
        dataset.delete()


@pytest.fixture()
def configured_project_by_global_key(
    client: Client,
    rand_gen,
    data_row_json_by_media_type,
    request: FixtureRequest,
    normalized_ontology_by_media_type,
    export_v2_test_helpers,
    teardown_helpers,
):
    """Does the same thing as configured project but with global keys focus."""

    media_type = getattr(request, "param", MediaType.Image)
    dataset = None

    if (
        media_type == MediaType.LLMPromptCreation
        or media_type == MediaType.LLMPromptResponseCreation
    ):
        project, ontology = _create_prompt_response_project(
            client,
            rand_gen,
            media_type,
            normalized_ontology_by_media_type,
            export_v2_test_helpers,
        )
    elif media_type == OntologyKind.ResponseCreation:
        project, ontology, dataset = _create_response_creation_project(
            client,
            rand_gen,
            data_row_json_by_media_type,
            media_type,
            normalized_ontology_by_media_type,
        )
    elif media_type == OntologyKind.ModelEvaluation:
        project, ontology, dataset = _create_offline_mmc_project(
            client,
            rand_gen,
            data_row_json_by_media_type[media_type],
            normalized_ontology_by_media_type[media_type],
        )
    else:
        project, ontology, dataset = _create_project(
            client,
            rand_gen,
            data_row_json_by_media_type,
            media_type,
            normalized_ontology_by_media_type,
        )

    yield project

    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)

    if dataset:
        dataset.delete()


@pytest.fixture(scope="module")
def module_project(
    client: Client,
    rand_gen,
    data_row_json_by_media_type,
    request: FixtureRequest,
    normalized_ontology_by_media_type,
    module_teardown_helpers,
):
    """Generates a image project that scopes to the test module(file). Used to reduce api calls."""

    media_type = getattr(request, "param", MediaType.Image)
    media_type = getattr(request, "param", MediaType.Image)
    dataset = None

    if (
        media_type == MediaType.LLMPromptCreation
        or media_type == MediaType.LLMPromptResponseCreation
    ):
        project, ontology = _create_prompt_response_project(
            client, rand_gen, media_type, normalized_ontology_by_media_type
        )
    elif media_type == OntologyKind.ResponseCreation:
        project, ontology, dataset = _create_response_creation_project(
            client,
            rand_gen,
            data_row_json_by_media_type,
            media_type,
            normalized_ontology_by_media_type,
        )
    else:
        project, ontology, dataset = _create_project(
            client,
            rand_gen,
            data_row_json_by_media_type,
            media_type,
            normalized_ontology_by_media_type,
        )

    yield project

    module_teardown_helpers.teardown_project_labels_ontology_feature_schemas(
        project
    )

    if dataset:
        dataset.delete()


@pytest.fixture
def prediction_id_mapping(request, normalized_ontology_by_media_type):
    """Creates the base of annotation based on tools inside project ontology. We would want only annotations supported for the MediaType of the ontology and project. Annotations are generated for each data row created later be combined inside the test file. This serves as the base fixture for all the interference (annotations) fixture. This fixtures supports a few strategies:

    Integration test:
        configured_project: generates data rows with data row id focus.
        configured_project_by_global_key: generates data rows with global key focus.
        module_configured_project: configured project but scoped to test module.

    Unit tests
        Individuals can supply hard-coded data row ids or global keys without configured a project must include a media type fixture to get the appropriate annotations.

    Each strategy provides a few items.

        Labelbox Project (unit testing strategies do not make api calls so will have None for project)
        Data row identifiers (ids the annotation uses)
        Ontology: normalized ontology
    """

    if "configured_project" in request.fixturenames:
        project = request.getfixturevalue("configured_project")
        data_row_identifiers = [
            {"id": data_row_id} for data_row_id in project.data_row_ids
        ]
        ontology = project.ontology().normalized

    elif "configured_project_by_global_key" in request.fixturenames:
        project = request.getfixturevalue("configured_project_by_global_key")
        data_row_identifiers = [
            {"globalKey": global_key} for global_key in project.global_keys
        ]
        ontology = project.ontology().normalized

    elif "module_project" in request.fixturenames:
        project = request.getfixturevalue("module_project")
        data_row_identifiers = [
            {"id": data_row_id} for data_row_id in project.data_row_ids
        ]
        ontology = project.ontology().normalized

    elif "hardcoded_datarow_id" in request.fixturenames:
        if "media_type" not in request.fixturenames:
            raise Exception("Please include a 'media_type' fixture")
        project = None
        media_type = request.getfixturevalue("media_type")
        ontology = normalized_ontology_by_media_type[media_type]
        data_row_identifiers = [
            {"id": request.getfixturevalue("hardcoded_datarow_id")()}
        ]

    elif "hardcoded_global_key" in request.fixturenames:
        if "media_type" not in request.fixturenames:
            raise Exception("Please include a 'media_type' fixture")
        project = None
        media_type = request.getfixturevalue("media_type")
        ontology = normalized_ontology_by_media_type[media_type]
        data_row_identifiers = [
            {"globalKey": request.getfixturevalue("hardcoded_global_key")()}
        ]

    # Used for tests that need access to every ontology
    else:
        project = None
        media_type = None
        ontology = normalized_ontology_by_media_type["all"]
        data_row_identifiers = [{"id": "ck8q9q9qj00003g5z3q1q9q9q"}]

    base_annotations = []
    for data_row_identifier in data_row_identifiers:
        base_annotation = {}
        for feature in ontology["tools"] + ontology["classifications"]:
            if "tool" in feature:
                feature_type = (
                    feature["tool"]
                    if feature["classifications"] == []
                    else f"{feature['tool']}_nested"
                )  # tool vs nested classification tool
            else:
                feature_type = (
                    feature["type"]
                    if "scope" not in feature
                    else f"{feature['type']}_{feature['scope']}"
                )  # checklist vs indexed checklist

            base_annotation[feature_type] = {
                "uuid": str(uuid.uuid4()),
                "name": feature["name"],
                "tool": feature,
                "dataRow": data_row_identifier,
            }

        base_annotations.append(base_annotation)
    return base_annotations


@pytest.fixture
def mmc_example_data_row_message_ids(mmc_data_row_url: str):
    data_row_content = requests.get(mmc_data_row_url).json()

    human_id = next(
        actor_id
        for actor_id, actor_metadata in data_row_content["actors"].items()
        if actor_metadata["role"] == "human"
    )

    return {
        message_id: [
            {
                "id": child_msg_id,
                "model_config_name": data_row_content["actors"][
                    data_row_content["messages"][child_msg_id]["actorId"]
                ]["metadata"]["modelConfigName"],
            }
            for child_msg_id in message_metadata["childMessageIds"]
        ]
        for message_id, message_metadata in data_row_content["messages"].items()
        if message_metadata["actorId"] == human_id
    }


# Each inference represents a feature type that adds to the base annotation created with prediction_id_mapping
@pytest.fixture
def polygon_inference(prediction_id_mapping):
    polygons = []
    for feature in prediction_id_mapping:
        if "polygon" not in feature:
            continue
        polygon = feature["polygon"].copy()
        polygon.update(
            {
                "polygon": [
                    {"x": 147.692, "y": 118.154},
                    {"x": 142.769, "y": 104.923},
                    {"x": 57.846, "y": 118.769},
                    {"x": 28.308, "y": 169.846},
                ]
            }
        )
        del polygon["tool"]
        polygons.append(polygon)
    return polygons


@pytest.fixture
def rectangle_inference(prediction_id_mapping):
    rectangles = []
    for feature in prediction_id_mapping:
        if "rectangle" not in feature:
            continue
        rectangle = feature["rectangle"].copy()
        rectangle.update(
            {
                "bbox": {"top": 48, "left": 58, "height": 65, "width": 12},
            }
        )
        del rectangle["tool"]
        rectangles.append(rectangle)
    return rectangles


@pytest.fixture
def rectangle_inference_with_confidence(prediction_id_mapping):
    rectangles = []
    for feature in prediction_id_mapping:
        if "rectangle_nested" not in feature:
            continue
        rectangle = feature["rectangle_nested"].copy()
        rectangle.update(
            {
                "bbox": {"top": 48, "left": 58, "height": 65, "width": 12},
                "classifications": [
                    {
                        "name": rectangle["tool"]["classifications"][0]["name"],
                        "answer": {
                            "name": rectangle["tool"]["classifications"][0][
                                "options"
                            ][0]["value"],
                            "classifications": [
                                {
                                    "name": rectangle["tool"][
                                        "classifications"
                                    ][0]["options"][0]["options"][1]["name"],
                                    "answer": "nested answer",
                                }
                            ],
                        },
                    }
                ],
            }
        )

        rectangle.update({"confidence": 0.9})
        rectangle["classifications"][0]["answer"]["confidence"] = 0.8
        rectangle["classifications"][0]["answer"]["classifications"][0][
            "confidence"
        ] = 0.7

        del rectangle["tool"]
        rectangles.append(rectangle)
    return rectangles


@pytest.fixture
def rectangle_inference_document(rectangle_inference):
    rectangles = []
    for feature in rectangle_inference:
        rectangle = feature.copy()
        rectangle.update({"page": 1, "unit": "POINTS"})
        rectangles.append(rectangle)
    return rectangles


@pytest.fixture
def line_inference(prediction_id_mapping):
    lines = []
    for feature in prediction_id_mapping:
        if "line" not in feature:
            continue
        line = feature["line"].copy()
        line.update(
            {
                "line": [
                    {"x": 147.692, "y": 118.154},
                    {"x": 150.692, "y": 160.154},
                ]
            }
        )
        del line["tool"]
        lines.append(line)
    return lines


@pytest.fixture
def point_inference(prediction_id_mapping):
    points = []
    for feature in prediction_id_mapping:
        if "point" not in feature:
            continue
        point = feature["point"].copy()
        point.update({"point": {"x": 147.692, "y": 118.154}})
        del point["tool"]
        points.append(point)
    return points


@pytest.fixture
def entity_inference(prediction_id_mapping):
    named_entities = []
    for feature in prediction_id_mapping:
        if "named-entity" not in feature:
            continue
        entity = feature["named-entity"].copy()
        entity.update({"location": {"start": 112, "end": 128}})
        del entity["tool"]
        named_entities.append(entity)
    return named_entities


@pytest.fixture
def entity_inference_index(prediction_id_mapping):
    named_entities = []
    for feature in prediction_id_mapping:
        if "named-entity" not in feature:
            continue
        entity = feature["named-entity"].copy()
        entity.update(
            {
                "location": {"start": 0, "end": 8},
                "messageId": "0",
            }
        )
        del entity["tool"]
        named_entities.append(entity)
    return named_entities


@pytest.fixture
def entity_inference_document(prediction_id_mapping):
    named_entities = []
    for feature in prediction_id_mapping:
        if "named-entity" not in feature:
            continue
        entity = feature["named-entity"].copy()
        document_selections = {
            "textSelections": [
                {
                    "tokenIds": [
                        "3f984bf3-1d61-44f5-b59a-9658a2e3440f",
                        "3bf00b56-ff12-4e52-8cc1-08dbddb3c3b8",
                        "6e1c3420-d4b7-4c5a-8fd6-ead43bf73d80",
                        "87a43d32-af76-4a1d-b262-5c5f4d5ace3a",
                        "e8606e8a-dfd9-4c49-a635-ad5c879c75d0",
                        "67c7c19e-4654-425d-bf17-2adb8cf02c30",
                        "149c5e80-3e07-49a7-ab2d-29ddfe6a38fa",
                        "b0e94071-2187-461e-8e76-96c58738a52c",
                    ],
                    "groupId": "2f4336f4-a07e-4e0a-a9e1-5629b03b719b",
                    "page": 1,
                }
            ]
        }
        entity.update(document_selections)
        del entity["tool"]
        named_entities.append(entity)
    return named_entities


@pytest.fixture
def segmentation_inference(prediction_id_mapping):
    superpixel_masks = []
    for feature in prediction_id_mapping:
        if "superpixel" not in feature:
            continue
        segmentation = feature["superpixel"].copy()
        segmentation.update(
            {
                "mask": {
                    "instanceURI": "https://storage.googleapis.com/labelbox-datasets/image_sample_data/raster_seg.png",
                    "colorRGB": (255, 255, 255),
                }
            }
        )
        del segmentation["tool"]
        superpixel_masks.append(segmentation)
    return superpixel_masks


@pytest.fixture
def segmentation_inference_rle(prediction_id_mapping):
    superpixel_masks = []
    for feature in prediction_id_mapping:
        if "superpixel" not in feature:
            continue
        segmentation = feature["superpixel"].copy()
        segmentation.update(
            {
                "uuid": str(uuid.uuid4()),
                "mask": {"size": [10, 10], "counts": [1, 0, 10, 100]},
            }
        )
        del segmentation["tool"]
        superpixel_masks.append(segmentation)
    return superpixel_masks


@pytest.fixture
def segmentation_inference_png(prediction_id_mapping):
    superpixel_masks = []
    for feature in prediction_id_mapping:
        if "superpixel" not in feature:
            continue
        segmentation = feature["superpixel"].copy()
        segmentation.update(
            {
                "uuid": str(uuid.uuid4()),
                "mask": {
                    "png": "somedata",
                },
            }
        )
        del segmentation["tool"]
        superpixel_masks.append(segmentation)
    return superpixel_masks


@pytest.fixture
def checklist_inference(prediction_id_mapping):
    checklists = []
    for feature in prediction_id_mapping:
        if "checklist" not in feature:
            continue
        checklist = feature["checklist"].copy()
        checklist.update(
            {
                "answers": [
                    {"name": "first_checklist_answer"},
                    {"name": "second_checklist_answer"},
                ]
            }
        )
        del checklist["tool"]
        checklists.append(checklist)
    return checklists


@pytest.fixture
def checklist_inference_index(prediction_id_mapping):
    checklists = []
    for feature in prediction_id_mapping:
        if "checklist_index" not in feature:
            return None
        checklist = feature["checklist_index"].copy()
        checklist.update(
            {
                "answers": [
                    {"name": "first_checklist_answer"},
                    {"name": "second_checklist_answer"},
                ],
                "messageId": "0",
            }
        )
        del checklist["tool"]
        checklists.append(checklist)
    return checklists


@pytest.fixture
def checklist_inference_index_mmc(
    prediction_id_mapping, mmc_example_data_row_message_ids
):
    checklists = []
    for feature in prediction_id_mapping:
        if "checklist_index" not in feature:
            return None
        checklist = feature["checklist_index"].copy()
        checklist.update(
            {
                "answers": [
                    {"name": "first_checklist_answer"},
                    {"name": "second_checklist_answer"},
                ],
                "messageId": next(
                    iter(mmc_example_data_row_message_ids.keys())
                ),
            }
        )
        del checklist["tool"]
        checklists.append(checklist)
    return checklists


@pytest.fixture
def prompt_text_inference(prediction_id_mapping):
    prompt_texts = []
    for feature in prediction_id_mapping:
        if "prompt" not in feature:
            continue
        text = feature["prompt"].copy()
        text.update({"answer": "free form text..."})
        del text["tool"]
        prompt_texts.append(text)
    return prompt_texts


@pytest.fixture
def radio_response_inference(prediction_id_mapping):
    response_radios = []
    for feature in prediction_id_mapping:
        if "response-radio" not in feature:
            continue
        response_radio = feature["response-radio"].copy()
        response_radio.update(
            {
                "answer": {"name": "first_radio_answer"},
            }
        )
        del response_radio["tool"]
        response_radios.append(response_radio)
    return response_radios


@pytest.fixture
def radio_inference(prediction_id_mapping):
    radios = []
    for feature in prediction_id_mapping:
        if "radio" not in feature:
            continue
        radio = feature["radio"].copy()
        radio.update(
            {
                "answer": {"name": "first_radio_answer"},
            }
        )
        del radio["tool"]
        radios.append(radio)
    return radios


@pytest.fixture
def radio_inference_index_mmc(
    prediction_id_mapping, mmc_example_data_row_message_ids
):
    radios = []
    for feature in prediction_id_mapping:
        if "radio_index" not in feature:
            continue
        radio = feature["radio_index"].copy()
        radio.update(
            {
                "answer": {"name": "first_radio_answer"},
                "messageId": next(
                    iter(mmc_example_data_row_message_ids.keys())
                ),
            }
        )
        del radio["tool"]
        radios.append(radio)
    return radios


@pytest.fixture
def checklist_response_inference(prediction_id_mapping):
    response_checklists = []
    for feature in prediction_id_mapping:
        if "response-checklist" not in feature:
            continue
        response_checklist = feature["response-checklist"].copy()
        response_checklist.update(
            {
                "answer": [
                    {"name": "first_checklist_answer"},
                    {"name": "second_checklist_answer"},
                ]
            }
        )
        del response_checklist["tool"]
        response_checklists.append(response_checklist)
    return response_checklists


@pytest.fixture
def text_response_inference(prediction_id_mapping):
    response_texts = []
    for feature in prediction_id_mapping:
        if "response-text" not in feature:
            continue
        text = feature["response-text"].copy()
        text.update({"answer": "free form text..."})
        del text["tool"]
        response_texts.append(text)
    return response_texts


@pytest.fixture
def text_inference(prediction_id_mapping):
    texts = []
    for feature in prediction_id_mapping:
        if "text" not in feature:
            continue
        text = feature["text"].copy()
        text.update({"answer": "free form text..."})
        del text["tool"]
        texts.append(text)
    return texts


@pytest.fixture
def text_inference_with_confidence(text_inference):
    texts = []
    for feature in text_inference:
        text = feature.copy()
        text.update({"confidence": 0.9})
        texts.append(text)
    return texts


@pytest.fixture
def text_inference_index(prediction_id_mapping):
    texts = []
    for feature in prediction_id_mapping:
        if "text_index" not in feature:
            continue
        text = feature["text_index"].copy()
        text.update({"answer": "free form text...", "messageId": "0"})
        del text["tool"]
        texts.append(text)
    return texts


@pytest.fixture
def text_inference_index_mmc(
    prediction_id_mapping, mmc_example_data_row_message_ids
):
    texts = []
    for feature in prediction_id_mapping:
        if "text_index" not in feature:
            continue
        text = feature["text_index"].copy()
        text.update(
            {
                "answer": "free form text...",
                "messageId": next(
                    iter(mmc_example_data_row_message_ids.keys())
                ),
            }
        )
        del text["tool"]
        texts.append(text)
    return texts


@pytest.fixture
def video_checklist_inference(prediction_id_mapping):
    checklists = []
    for feature in prediction_id_mapping:
        if "checklist_index" not in feature:
            continue
        checklist = feature["checklist_index"].copy()
        checklist.update(
            {
                "answers": [
                    {"name": "first_checklist_answer"},
                    {"name": "second_checklist_answer"},
                ]
            }
        )

        checklist.update(
            {
                "frames": [
                    {
                        "start": 7,
                        "end": 13,
                    },
                    {
                        "start": 18,
                        "end": 19,
                    },
                ]
            }
        )
        del checklist["tool"]
        checklists.append(checklist)
    return checklists


@pytest.fixture
def message_single_selection_inference(
    prediction_id_mapping, mmc_example_data_row_message_ids
):
    some_parent_id, some_child_ids = next(
        iter(mmc_example_data_row_message_ids.items())
    )

    res = []
    for feature in prediction_id_mapping:
        if "message-single-selection" not in feature:
            continue
        selection = feature["message-single-selection"].copy()
        selection.update(
            {
                "messageEvaluationTask": {
                    "format": "message-single-selection",
                    "data": {
                        "messageId": some_child_ids[0]["id"],
                        "parentMessageId": some_parent_id,
                        "modelConfigName": some_child_ids[0][
                            "model_config_name"
                        ],
                    },
                }
            }
        )
        del selection["tool"]
        res.append(selection)

    return res


@pytest.fixture
def message_multi_selection_inference(
    prediction_id_mapping, mmc_example_data_row_message_ids
):
    some_parent_id, some_child_ids = next(
        iter(mmc_example_data_row_message_ids.items())
    )

    res = []
    for feature in prediction_id_mapping:
        if "message-multi-selection" not in feature:
            continue
        selection = feature["message-multi-selection"].copy()
        selection.update(
            {
                "messageEvaluationTask": {
                    "format": "message-multi-selection",
                    "data": {
                        "parentMessageId": some_parent_id,
                        "selectedMessages": [
                            {
                                "messageId": child_id["id"],
                                "modelConfigName": child_id[
                                    "model_config_name"
                                ],
                            }
                            for child_id in some_child_ids
                        ],
                    },
                }
            }
        )
        del selection["tool"]
        res.append(selection)

    return res


@pytest.fixture
def message_ranking_inference(
    prediction_id_mapping, mmc_example_data_row_message_ids
):
    some_parent_id, some_child_ids = next(
        iter(mmc_example_data_row_message_ids.items())
    )

    res = []
    for feature in prediction_id_mapping:
        if "message-ranking" not in feature:
            continue
        selection = feature["message-ranking"].copy()
        selection.update(
            {
                "messageEvaluationTask": {
                    "format": "message-ranking",
                    "data": {
                        "parentMessageId": some_parent_id,
                        "rankedMessages": [
                            {
                                "messageId": child_id["id"],
                                "modelConfigName": child_id[
                                    "model_config_name"
                                ],
                                "order": idx,
                            }
                            for idx, child_id in enumerate(
                                some_child_ids, start=1
                            )
                        ],
                    },
                }
            }
        )
        del selection["tool"]
        res.append(selection)

    return res


@pytest.fixture
def annotations_by_media_type(
    polygon_inference,
    rectangle_inference,
    rectangle_inference_document,
    line_inference,
    entity_inference,
    entity_inference_index,
    entity_inference_document,
    checklist_inference_index,
    text_inference_index,
    checklist_inference,
    text_inference,
    video_checklist_inference,
    prompt_text_inference,
    checklist_response_inference,
    radio_response_inference,
    text_response_inference,
    message_single_selection_inference,
    message_multi_selection_inference,
    message_ranking_inference,
    checklist_inference_index_mmc,
    radio_inference,
    radio_inference_index_mmc,
    text_inference_index_mmc,
):
    return {
        MediaType.Audio: [checklist_inference, text_inference],
        MediaType.Conversational: [
            checklist_inference_index,
            text_inference_index,
            entity_inference_index,
        ],
        MediaType.Document: [
            entity_inference_document,
            checklist_inference,
            text_inference,
            rectangle_inference_document,
        ],
        MediaType.Html: [text_inference, checklist_inference],
        MediaType.Image: [
            polygon_inference,
            rectangle_inference,
            line_inference,
            checklist_inference,
            text_inference,
        ],
        MediaType.Text: [checklist_inference, text_inference, entity_inference],
        MediaType.Video: [video_checklist_inference],
        MediaType.LLMPromptResponseCreation: [
            prompt_text_inference,
            text_response_inference,
            checklist_response_inference,
            radio_response_inference,
        ],
        MediaType.LLMPromptCreation: [prompt_text_inference],
        OntologyKind.ResponseCreation: [
            text_response_inference,
            checklist_response_inference,
            radio_response_inference,
        ],
        OntologyKind.ModelEvaluation: [
            message_single_selection_inference,
            message_multi_selection_inference,
            message_ranking_inference,
            radio_inference,
            checklist_inference,
            text_inference,
            radio_inference_index_mmc,
            checklist_inference_index_mmc,
            text_inference_index_mmc,
        ],
    }


@pytest.fixture
def model_run_predictions(
    polygon_inference, rectangle_inference, line_inference
):
    # Not supporting mask since there isn't a signed url representing a seg mask to upload
    return polygon_inference + rectangle_inference + line_inference


@pytest.fixture
def object_predictions(
    polygon_inference,
    rectangle_inference,
    line_inference,
    entity_inference,
    segmentation_inference,
):
    return (
        polygon_inference
        + rectangle_inference
        + line_inference
        + entity_inference
        + segmentation_inference
    )


@pytest.fixture
def object_predictions_for_annotation_import(
    polygon_inference,
    rectangle_inference,
    line_inference,
    segmentation_inference,
):
    return (
        polygon_inference
        + rectangle_inference
        + line_inference
        + segmentation_inference
    )


@pytest.fixture
def classification_predictions(checklist_inference, text_inference):
    return checklist_inference + text_inference


@pytest.fixture
def predictions(object_predictions, classification_predictions):
    return object_predictions + classification_predictions


# Can only have confidence predictions supported by media type of project
@pytest.fixture
def predictions_with_confidence(rectangle_inference_with_confidence):
    return rectangle_inference_with_confidence


@pytest.fixture
def model(client, rand_gen, configured_project):
    ontology = configured_project.ontology()
    data = {"name": rand_gen(str), "ontology_id": ontology.uid}
    model = client.create_model(data["name"], data["ontology_id"])
    yield model
    try:
        model.delete()
    except:
        # Already was deleted by the test
        pass


@pytest.fixture
def model_run(rand_gen, model):
    name = rand_gen(str)
    model_run = model.create_model_run(name)
    yield model_run
    try:
        model_run.delete()
    except:
        # Already was deleted by the test
        pass


@pytest.fixture
def model_run_with_training_metadata(rand_gen, model):
    name = rand_gen(str)
    training_metadata = {"batch_size": 1000}
    model_run = model.create_model_run(name, training_metadata)
    yield model_run
    try:
        model_run.delete()
    except:
        # Already was deleted by the test
        pass


@pytest.fixture
def model_run_with_data_rows(
    client,
    configured_project,
    model_run_predictions,
    model_run,
    wait_for_label_processing,
):
    use_data_row_ids = [p["dataRow"]["id"] for p in model_run_predictions]
    model_run.upsert_data_rows(use_data_row_ids)

    upload_task = LabelImport.create_from_objects(
        client,
        configured_project.uid,
        f"label-import-{uuid.uuid4()}",
        model_run_predictions,
    )
    upload_task.wait_until_done()
    assert (
        upload_task.state == AnnotationImportState.FINISHED
    ), "Label Import did not finish"
    assert (
        len(upload_task.errors) == 0
    ), f"Label Import {upload_task.name} failed with errors {upload_task.errors}"
    labels = wait_for_label_processing(configured_project)
    label_ids = [label.uid for label in labels]
    model_run.upsert_labels(label_ids)
    yield model_run
    model_run.delete()


@pytest.fixture
def model_run_with_all_project_labels(
    client,
    configured_project,
    model_run_predictions,
    model_run: ModelRun,
    wait_for_label_processing,
):
    use_data_row_ids = list(
        set([p["dataRow"]["id"] for p in model_run_predictions])
    )

    model_run.upsert_data_rows(use_data_row_ids)

    upload_task = LabelImport.create_from_objects(
        client,
        configured_project.uid,
        f"label-import-{uuid.uuid4()}",
        model_run_predictions,
    )
    upload_task.wait_until_done()
    assert (
        upload_task.state == AnnotationImportState.FINISHED
    ), "Label Import did not finish"
    assert (
        len(upload_task.errors) == 0
    ), f"Label Import {upload_task.name} failed with errors {upload_task.errors}"
    labels = wait_for_label_processing(configured_project)
    label_ids = [label.uid for label in labels]
    model_run.upsert_labels(label_ids)
    yield model_run
    model_run.delete()


class AnnotationImportTestHelpers:
    @classmethod
    def assert_file_content(cls, url: str, predictions):
        response = requests.get(url)
        predictions = cls._convert_to_plain_object(predictions)
        assert parser.loads(response.text) == predictions

    @staticmethod
    def check_running_state(req, name, url=None):
        assert req.name == name
        if url is not None:
            assert req.input_file_url == url
        assert req.error_file_url is None
        assert req.status_file_url is None
        assert req.state == AnnotationImportState.RUNNING

    @staticmethod
    def download_and_assert_status(status_file_url):
        response = requests.get(status_file_url)
        assert response.status_code == 200
        for line in parser.loads(response.content):
            status = line["status"]
            assert status.upper() == "SUCCESS"

    @staticmethod
    def _convert_to_plain_object(obj):
        """Some Python objects e.g. tuples can't be compared with JSON serialized data, serialize to JSON and deserialize to get plain objects"""
        json_str = parser.dumps(obj)
        return parser.loads(json_str)


@pytest.fixture
def annotation_import_test_helpers() -> Type[AnnotationImportTestHelpers]:
    return AnnotationImportTestHelpers()


@pytest.fixture()
def expected_export_v2_image():
    exported_annotations = {
        "objects": [
            {
                "name": "polygon",
                "value": "polygon",
                "annotation_kind": "ImagePolygon",
                "classifications": [],
                "polygon": [
                    {"x": 147.692, "y": 118.154},
                    {"x": 142.769, "y": 104.923},
                    {"x": 57.846, "y": 118.769},
                    {"x": 28.308, "y": 169.846},
                    {"x": 147.692, "y": 118.154},
                ],
            },
            {
                "name": "bbox",
                "value": "bbox",
                "annotation_kind": "ImageBoundingBox",
                "classifications": [],
                "bounding_box": {
                    "top": 48.0,
                    "left": 58.0,
                    "height": 65.0,
                    "width": 12.0,
                },
            },
            {
                "name": "polyline",
                "value": "polyline",
                "annotation_kind": "ImagePolyline",
                "classifications": [],
                "line": [
                    {"x": 147.692, "y": 118.154},
                    {"x": 150.692, "y": 160.154},
                ],
            },
        ],
        "classifications": [
            {
                "name": "checklist",
                "value": "checklist",
                "checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
            {
                "name": "text",
                "value": "text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
        ],
        "relationships": [],
    }

    return exported_annotations


@pytest.fixture()
def expected_export_v2_audio():
    expected_annotations = {
        "classifications": [
            {
                "name": "checklist",
                "value": "checklist",
                "checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
            {
                "name": "text",
                "value": "text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
        ],
        "segments": {},
        "timestamp": {},
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_html():
    expected_annotations = {
        "objects": [],
        "classifications": [
            {
                "name": "text",
                "value": "text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
            {
                "name": "checklist",
                "value": "checklist",
                "checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
        ],
        "relationships": [],
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_text():
    expected_annotations = {
        "objects": [
            {
                "name": "named-entity",
                "value": "named_entity",
                "annotation_kind": "TextEntity",
                "classifications": [],
                "location": {
                    "start": 112,
                    "end": 128,
                    "token": "research suggests",
                },
            }
        ],
        "classifications": [
            {
                "name": "checklist",
                "value": "checklist",
                "checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
            {
                "name": "text",
                "value": "text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
        ],
        "relationships": [],
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_video():
    expected_annotations = {
        "classifications": [],
        "frames": {
            "7": {
                "objects": {},
                "classifications": [
                    {
                        "name": "checklist_index",
                        "value": "checklist_index",
                        "checklist_answers": [
                            {
                                "name": "first_checklist_answer",
                                "value": "first_checklist_answer",
                                "classifications": [],
                            },
                            {
                                "name": "second_checklist_answer",
                                "value": "second_checklist_answer",
                                "classifications": [],
                            },
                        ],
                    }
                ],
            },
            "13": {
                "objects": {},
                "classifications": [
                    {
                        "name": "checklist_index",
                        "value": "checklist_index",
                        "checklist_answers": [
                            {
                                "name": "first_checklist_answer",
                                "value": "first_checklist_answer",
                                "classifications": [],
                            },
                            {
                                "name": "second_checklist_answer",
                                "value": "second_checklist_answer",
                                "classifications": [],
                            },
                        ],
                    }
                ],
            },
            "18": {
                "objects": {},
                "classifications": [
                    {
                        "name": "checklist_index",
                        "value": "checklist_index",
                        "checklist_answers": [
                            {
                                "name": "first_checklist_answer",
                                "value": "first_checklist_answer",
                                "classifications": [],
                            },
                            {
                                "name": "second_checklist_answer",
                                "value": "second_checklist_answer",
                                "classifications": [],
                            },
                        ],
                    }
                ],
            },
            "19": {
                "objects": {},
                "classifications": [
                    {
                        "name": "checklist_index",
                        "value": "checklist_index",
                        "checklist_answers": [
                            {
                                "name": "first_checklist_answer",
                                "value": "first_checklist_answer",
                                "classifications": [],
                            },
                            {
                                "name": "second_checklist_answer",
                                "value": "second_checklist_answer",
                                "classifications": [],
                            },
                        ],
                    }
                ],
            },
        },
        "segments": {
            "<cuid>": [[7, 13], [18, 19]],
        },
        "key_frame_feature_map": {
            "<cuid>": [7, 13, 18, 19],
        },
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_conversation():
    expected_annotations = {
        "objects": [
            {
                "name": "named-entity",
                "value": "named_entity",
                "annotation_kind": "ConversationalTextEntity",
                "classifications": [],
                "conversational_location": {
                    "message_id": "0",
                    "location": {"start": 0, "end": 8},
                },
            }
        ],
        "classifications": [
            {
                "name": "checklist_index",
                "value": "checklist_index",
                "message_id": "0",
                "conversational_checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
            {
                "name": "text_index",
                "value": "text_index",
                "message_id": "0",
                "conversational_text_answer": {"content": "free form text..."},
            },
        ],
        "relationships": [],
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_document():
    expected_annotations = {
        "objects": [
            {
                "name": "named-entity",
                "value": "named_entity",
                "annotation_kind": "DocumentEntityToken",
                "classifications": [],
                "location": {
                    "groups": [
                        {
                            "id": "2f4336f4-a07e-4e0a-a9e1-5629b03b719b",
                            "page_number": 1,
                            "tokens": [
                                "3f984bf3-1d61-44f5-b59a-9658a2e3440f",
                                "3bf00b56-ff12-4e52-8cc1-08dbddb3c3b8",
                                "6e1c3420-d4b7-4c5a-8fd6-ead43bf73d80",
                                "87a43d32-af76-4a1d-b262-5c5f4d5ace3a",
                                "e8606e8a-dfd9-4c49-a635-ad5c879c75d0",
                                "67c7c19e-4654-425d-bf17-2adb8cf02c30",
                                "149c5e80-3e07-49a7-ab2d-29ddfe6a38fa",
                                "b0e94071-2187-461e-8e76-96c58738a52c",
                            ],
                            "text": "Metal-insulator (MI) transitions have been one of the",
                        }
                    ]
                },
            },
            {
                "name": "bbox",
                "value": "bbox",
                "annotation_kind": "DocumentBoundingBox",
                "classifications": [],
                "page_number": 1,
                "bounding_box": {
                    "top": 48.0,
                    "left": 58.0,
                    "height": 65.0,
                    "width": 12.0,
                    "unit": "points",
                },
                "page_dimensions": {
                    "height": 792.0,
                    "width": 612.0,
                    "unit": "points",
                },
            },
        ],
        "classifications": [
            {
                "name": "checklist",
                "value": "checklist",
                "checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
            {
                "name": "text",
                "value": "text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
        ],
        "relationships": [],
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_llm_prompt_response_creation():
    expected_annotations = {
        "objects": [],
        "classifications": [
            {
                "name": "prompt-text",
                "value": "prompt-text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
            {
                "name": "response-text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
                "value": "response-text",
            },
            {
                "checklist_answers": [
                    {
                        "classifications": [],
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                    },
                    {
                        "classifications": [],
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                    },
                ],
                "name": "checklist-response",
                "value": "checklist-response",
            },
            {
                "name": "radio-response",
                "radio_answer": {
                    "classifications": [],
                    "name": "first_radio_answer",
                    "value": "first_radio_answer",
                },
                "value": "radio-response",
            },
        ],
        "relationships": [],
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_llm_prompt_creation():
    expected_annotations = {
        "objects": [],
        "classifications": [
            {
                "name": "prompt-text",
                "value": "prompt-text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
        ],
        "relationships": [],
    }
    return expected_annotations


@pytest.fixture()
def expected_export_v2_llm_response_creation():
    expected_annotations = {
        "objects": [],
        "relationships": [],
        "classifications": [
            {
                "name": "response-text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
                "value": "response-text",
            },
            {
                "checklist_answers": [
                    {
                        "classifications": [],
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                    },
                    {
                        "classifications": [],
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                    },
                ],
                "name": "checklist-response",
                "value": "checklist-response",
            },
            {
                "name": "radio-response",
                "radio_answer": {
                    "classifications": [],
                    "name": "first_radio_answer",
                    "value": "first_radio_answer",
                },
                "value": "radio-response",
            },
        ],
    }
    return expected_annotations


@pytest.fixture
def expected_exports_v2_mmc(mmc_example_data_row_message_ids):
    some_parent_id, some_child_ids = next(
        iter(mmc_example_data_row_message_ids.items())
    )

    return {
        "objects": [
            {
                "name": "message-single-selection",
                "annotation_kind": "MessageSingleSelection",
                "classifications": [],
                "selected_message": {
                    "message_id": some_child_ids[0]["id"],
                    "model_config_name": some_child_ids[0]["model_config_name"],
                    "parent_message_id": some_parent_id,
                },
            },
            {
                "name": "message-multi-selection",
                "annotation_kind": "MessageMultiSelection",
                "classifications": [],
                "selected_messages": {
                    "messages": [
                        {
                            "message_id": child_id["id"],
                            "model_config_name": child_id["model_config_name"],
                        }
                        for child_id in some_child_ids
                    ],
                    "parent_message_id": some_parent_id,
                },
            },
            {
                "name": "message-ranking",
                "annotation_kind": "MessageRanking",
                "classifications": [],
                "ranked_messages": {
                    "ranked_messages": [
                        {
                            "message_id": child_id["id"],
                            "model_config_name": child_id["model_config_name"],
                            "order": idx,
                        }
                        for idx, child_id in enumerate(some_child_ids, start=1)
                    ],
                    "parent_message_id": some_parent_id,
                },
            },
        ],
        "classifications": [
            {
                "name": "radio",
                "value": "radio",
                "radio_answer": {
                    "name": "first_radio_answer",
                    "value": "first_radio_answer",
                    "classifications": [],
                },
            },
            {
                "name": "checklist",
                "value": "checklist",
                "checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
            {
                "name": "text",
                "value": "text",
                "text_answer": {
                    "content": "free form text...",
                    "classifications": [],
                },
            },
            {
                "name": "radio_index",
                "value": "radio_index",
                "message_id": some_parent_id,
                "conversational_radio_answer": {
                    "name": "first_radio_answer",
                    "value": "first_radio_answer",
                    "classifications": [],
                },
            },
            {
                "name": "checklist_index",
                "value": "checklist_index",
                "message_id": some_parent_id,
                "conversational_checklist_answers": [
                    {
                        "name": "first_checklist_answer",
                        "value": "first_checklist_answer",
                        "classifications": [],
                    },
                    {
                        "name": "second_checklist_answer",
                        "value": "second_checklist_answer",
                        "classifications": [],
                    },
                ],
            },
            {
                "name": "text_index",
                "value": "text_index",
                "message_id": some_parent_id,
                "conversational_text_answer": {"content": "free form text..."},
            },
        ],
        "relationships": [],
    }


@pytest.fixture
def exports_v2_by_media_type(
    expected_export_v2_image,
    expected_export_v2_audio,
    expected_export_v2_html,
    expected_export_v2_text,
    expected_export_v2_video,
    expected_export_v2_conversation,
    expected_export_v2_document,
    expected_export_v2_llm_prompt_response_creation,
    expected_export_v2_llm_prompt_creation,
    expected_export_v2_llm_response_creation,
    expected_exports_v2_mmc,
):
    return {
        MediaType.Image: expected_export_v2_image,
        MediaType.Audio: expected_export_v2_audio,
        MediaType.Html: expected_export_v2_html,
        MediaType.Text: expected_export_v2_text,
        MediaType.Video: expected_export_v2_video,
        MediaType.Conversational: expected_export_v2_conversation,
        MediaType.Document: expected_export_v2_document,
        MediaType.LLMPromptResponseCreation: expected_export_v2_llm_prompt_response_creation,
        MediaType.LLMPromptCreation: expected_export_v2_llm_prompt_creation,
        OntologyKind.ResponseCreation: expected_export_v2_llm_response_creation,
        OntologyKind.ModelEvaluation: expected_exports_v2_mmc,
    }


class Helpers:
    @staticmethod
    def remove_keys_recursive(d, keys):
        for k in keys:
            if k in d:
                del d[k]
        for k, v in d.items():
            if isinstance(v, dict):
                Helpers.remove_keys_recursive(v, keys)
            elif isinstance(v, list):
                for i in v:
                    if isinstance(i, dict):
                        Helpers.remove_keys_recursive(i, keys)

    @staticmethod
    # NOTE this uses quite a primitive check for cuids but I do not think it is worth coming up with a better one
    # Also this function is NOT written with performance in mind, good for small to mid size dicts like we have in our test
    def rename_cuid_key_recursive(d):
        new_key = "<cuid>"
        for k in list(d.keys()):
            if len(k) == 25 and not k.isalpha():  # primitive check for cuid
                d[new_key] = d.pop(k)
        for k, v in d.items():
            if isinstance(v, dict):
                Helpers.rename_cuid_key_recursive(v)
            elif isinstance(v, list):
                for i in v:
                    if isinstance(i, dict):
                        Helpers.rename_cuid_key_recursive(i)

    @staticmethod
    def set_project_media_type_from_data_type(project, data_type_class):
        def to_pascal_case(name: str) -> str:
            return "".join([word.capitalize() for word in name.split("_")])

        data_type_string = data_type_class.__name__[:-4].lower()
        media_type = to_pascal_case(data_type_string)
        if media_type == "Conversational":
            media_type = "Conversational"
        elif media_type == "Llmpromptcreation":
            media_type = "LLMPromptCreation"
        elif media_type == "Llmpromptresponsecreation":
            media_type = "LLMPromptResponseCreation"
        elif media_type == "Llmresponsecreation":
            media_type = "Text"
        elif media_type == "Genericdatarow":
            media_type = "Image"
        project.update(media_type=MediaType[media_type])

    @staticmethod
    def find_data_row_filter(data_row):
        return lambda dr: dr["data_row"]["id"] == data_row.uid


@pytest.fixture
def helpers():
    return Helpers



import json
import uuid
from unittest.mock import patch

import pytest

from labelbox.schema.asset_attachment import AttachmentType
from labelbox.schema.identifiable import UniqueId, GlobalKey


class TestDataRowUpsert:
    @pytest.fixture
    def all_inclusive_data_row(self, dataset, image_url):
        dr = dataset.create_data_row(
            row_data=image_url,
            external_id="ex1",
            global_key=str(uuid.uuid4()),
            metadata_fields=[
                {"name": "tag", "value": "tag_string"},
                {"name": "split", "value": "train"},
            ],
            attachments=[
                {"type": "RAW_TEXT", "name": "att1", "value": "test1"},
                {
                    "type": "IMAGE",
                    "name": "att2",
                    "value": "https://storage.googleapis.com/labelbox-sample-datasets/Docs/disease_attachment.jpeg",
                },
                {
                    "type": "PDF_URL",
                    "name": "att3",
                    "value": "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf",
                },
            ],
        )
        return dr

    @pytest.mark.order(1)
    def test_create_data_row_with_auto_key(self, dataset, image_url):
        task = dataset.upsert_data_rows([{"row_data": image_url}])
        task.wait_till_done()
        assert len(list(dataset.data_rows())) == 1

    def test_create_data_row_with_upsert(self, client, dataset, image_url):
        gkey = str(uuid.uuid4())
        task = dataset.upsert_data_rows(
            [
                {
                    "row_data": image_url,
                    "global_key": gkey,
                    "external_id": "ex1",
                    "attachments": [
                        {
                            "type": AttachmentType.RAW_TEXT,
                            "name": "att1",
                            "value": "test1",
                        },
                        {
                            "type": AttachmentType.IMAGE,
                            "name": "att2",
                            "value": "https://storage.googleapis.com/labelbox-sample-datasets/Docs/disease_attachment.jpeg",
                        },
                        {
                            "type": AttachmentType.PDF_URL,
                            "name": "att3",
                            "value": "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf",
                        },
                    ],
                    "metadata": [
                        {"name": "tag", "value": "updated tag"},
                        {"name": "split", "value": "train"},
                    ],
                }
            ]
        )
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row_by_global_key(gkey)

        assert dr is not None
        assert dr.row_data == image_url
        assert dr.global_key == gkey
        assert dr.external_id == "ex1"

        attachments = list(dr.attachments())
        assert len(attachments) == 3
        assert attachments[0].attachment_name == "att1"
        assert attachments[0].attachment_type == AttachmentType.RAW_TEXT
        assert attachments[0].attachment_value == "test1"

        assert attachments[1].attachment_name == "att2"
        assert attachments[1].attachment_type == AttachmentType.IMAGE
        assert (
            attachments[1].attachment_value
            == "https://storage.googleapis.com/labelbox-sample-datasets/Docs/disease_attachment.jpeg"
        )

        assert attachments[2].attachment_name == "att3"
        assert attachments[2].attachment_type == AttachmentType.PDF_URL
        assert (
            attachments[2].attachment_value
            == "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf"
        )

        assert len(dr.metadata_fields) == 2
        assert dr.metadata_fields[0]["name"] == "tag"
        assert dr.metadata_fields[0]["value"] == "updated tag"
        assert dr.metadata_fields[1]["name"] == "split"
        assert dr.metadata_fields[1]["value"] == "train"

    def test_update_data_row_fields_with_upsert(
        self, client, dataset, image_url
    ):
        gkey = str(uuid.uuid4())
        dr = dataset.create_data_row(
            row_data=image_url, external_id="ex1", global_key=gkey
        )
        task = dataset.upsert_data_rows(
            [
                {
                    "key": UniqueId(dr.uid),
                    "external_id": "ex1_updated",
                    "global_key": f"{gkey}_updated",
                }
            ]
        )
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        assert dr.external_id == "ex1_updated"
        assert dr.global_key == f"{gkey}_updated"

    def test_update_data_row_fields_with_upsert_by_global_key(
        self, client, dataset, image_url
    ):
        gkey = str(uuid.uuid4())
        dr = dataset.create_data_row(
            row_data=image_url, external_id="ex1", global_key=gkey
        )
        task = dataset.upsert_data_rows(
            [
                {
                    "key": GlobalKey(dr.global_key),
                    "external_id": "ex1_updated",
                    "global_key": f"{gkey}_updated",
                }
            ]
        )
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        assert dr.external_id == "ex1_updated"
        assert dr.global_key == f"{gkey}_updated"

    def test_update_attachments_with_upsert(
        self, client, all_inclusive_data_row, dataset
    ):
        dr = all_inclusive_data_row
        task = dataset.upsert_data_rows(
            [
                {
                    "key": UniqueId(dr.uid),
                    "row_data": dr.row_data,
                    "attachments": [
                        {
                            "type": AttachmentType.RAW_TEXT,
                            "name": "att1",
                            "value": "test",
                        }
                    ],
                }
            ]
        )
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        attachments = list(dr.attachments())
        assert len(attachments) == 1
        assert attachments[0].attachment_name == "att1"

    def test_update_metadata_with_upsert(
        self, client, all_inclusive_data_row, dataset
    ):
        dr = all_inclusive_data_row
        task = dataset.upsert_data_rows(
            [
                {
                    "key": GlobalKey(dr.global_key),
                    "row_data": dr.row_data,
                    "metadata": [
                        {"name": "tag", "value": "updated tag"},
                        {"name": "split", "value": "train"},
                    ],
                }
            ]
        )
        task.wait_till_done()
        assert task.status == "COMPLETE"
        dr = client.get_data_row(dr.uid)
        assert dr is not None
        assert len(dr.metadata_fields) == 2
        assert dr.metadata_fields[0]["name"] == "tag"
        assert dr.metadata_fields[0]["value"] == "updated tag"
        assert dr.metadata_fields[1]["name"] == "split"
        assert dr.metadata_fields[1]["value"] == "train"

    def test_multiple_chunks(self, client, dataset, image_url):
        mocked_chunk_size = 300
        with patch(
            "labelbox.client.Client.upload_data", wraps=client.upload_data
        ) as spy_some_function:
            with patch(
                "labelbox.schema.dataset.UPSERT_CHUNK_SIZE_BYTES",
                new=mocked_chunk_size,
            ):
                task = dataset.upsert_data_rows(
                    [{"row_data": image_url} for i in range(10)]
                )
                task.wait_till_done()
                assert len(list(dataset.data_rows())) == 10
                assert (
                    spy_some_function.call_count == 11
                )  # one per each data row + manifest

                first_call_args, _ = spy_some_function.call_args_list[0]
                first_chunk_content = first_call_args[0]
                data = json.loads(first_chunk_content)
                """
                Each chunk but the last will have 3 items, the last will have 1.
                As the chunks are processed async, we cannot be sure of the order of the calls.
                """
                assert len(data) in {1, 3}

                last_call_args, _ = spy_some_function.call_args_list[-1]
                manifest_content = last_call_args[0].decode("utf-8")
                data = json.loads(manifest_content)
                assert data["source"] == "SDK"
                assert data["item_count"] == 10
                assert len(data["chunk_uris"]) == 10

    def test_upsert_embedded_row_data(self, dataset):
        pdf_url = "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/0801.3483.pdf"
        task = dataset.upsert_data_rows(
            [
                {
                    "row_data": {
                        "pdf_url": pdf_url,
                        "text_layer_url": "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/0801.3483-lb-textlayer.json",
                    },
                    "media_type": "PDF",
                }
            ]
        )
        task.wait_till_done()
        data_rows = list(dataset.data_rows())
        assert len(data_rows) == 1
        assert data_rows[0].row_data == pdf_url

    def test_upsert_duplicate_global_key_error(self, dataset, image_url):
        gkey = str(uuid.uuid4())
        task = dataset.upsert_data_rows(
            [
                {"row_data": image_url, "global_key": gkey},
                {"row_data": image_url, "global_key": gkey},
            ]
        )
        task.wait_till_done()
        assert task.status == "COMPLETE"
        assert task.errors is not None
        assert len(task.errors) == 1  # one data row was created, one failed
        assert f"Duplicate global key: '{gkey}'" in task.errors[0]["message"]

    def test_upsert_empty_items(self, dataset):
        items = [{"key": GlobalKey("foo")}]
        with pytest.raises(ValueError) as e_info:
            dataset.upsert_data_rows(items)
        e_info.match(r"The following items have an empty payload: \['foo'\]")



from typing import Dict, List, Tuple, Union

from labelbox.schema import ontology
from .annotation_types import (
    Text,
    Checklist,
    Radio,
    ClassificationAnnotation,
    ObjectAnnotation,
    Mask,
    Point,
    Line,
    Polygon,
    Rectangle,
    TextEntity,
)


def get_feature_schema_lookup(
    ontology_builder: ontology.OntologyBuilder,
) -> Tuple[Dict[str, str], Dict[str, str]]:
    tool_lookup = {}
    classification_lookup = {}

    def flatten_classification(classifications):
        for classification in classifications:
            if classification.feature_schema_id is None:
                raise ValueError(
                    f"feature_schema_id cannot be None for classification `{classification.name}`."
                )
            if isinstance(classification, ontology.Classification):
                classification_lookup[classification.name] = (
                    classification.feature_schema_id
                )
            elif isinstance(classification, ontology.Option):
                classification_lookup[classification.value] = (
                    classification.feature_schema_id
                )
            else:
                raise TypeError(
                    f"Unexpected type found in ontology. `{type(classification)}`"
                )
            flatten_classification(classification.options)

    for tool in ontology_builder.tools:
        if tool.feature_schema_id is None:
            raise ValueError(
                f"feature_schema_id cannot be None for tool `{tool.name}`."
            )
        tool_lookup[tool.name] = tool.feature_schema_id
        flatten_classification(tool.classifications)
    flatten_classification(ontology_builder.classifications)
    return tool_lookup, classification_lookup


def _get_options(
    annotation: ClassificationAnnotation,
    existing_options: List[ontology.Option],
):
    if isinstance(annotation.value, Radio):
        answers = [annotation.value.answer]
    elif isinstance(annotation.value, Text):
        return existing_options
    elif isinstance(annotation.value, (Checklist)):
        answers = annotation.value.answer
    else:
        raise TypeError(
            f"Expected one of Radio, Text, Checklist. Found {type(annotation.value)}"
        )

    option_names = {option.value for option in existing_options}
    for answer in answers:
        if answer.name not in option_names:
            existing_options.append(ontology.Option(value=answer.name))
            option_names.add(answer.name)
    return existing_options


def get_classifications(
    annotations: List[ClassificationAnnotation],
    existing_classifications: List[ontology.Classification],
) -> List[ontology.Classification]:
    existing_classifications = {
        classification.name: classification
        for classification in existing_classifications
    }
    for annotation in annotations:
        # If the classification exists then we just want to add options to it
        classification_feature = existing_classifications.get(annotation.name)
        if classification_feature:
            classification_feature.options = _get_options(
                annotation, classification_feature.options
            )
        elif annotation.name not in existing_classifications:
            existing_classifications[annotation.name] = ontology.Classification(
                class_type=classification_mapping(annotation),
                name=annotation.name,
                options=_get_options(annotation, []),
            )
    return list(existing_classifications.values())


def get_tools(
    annotations: List[ObjectAnnotation],
    existing_tools: List[ontology.Classification],
) -> List[ontology.Tool]:
    existing_tools = {tool.name: tool for tool in existing_tools}
    for annotation in annotations:
        if annotation.name in existing_tools:
            # We just want to update classifications
            existing_tools[
                annotation.name
            ].classifications = get_classifications(
                annotation.classifications,
                existing_tools[annotation.name].classifications,
            )
        else:
            existing_tools[annotation.name] = ontology.Tool(
                tool=tool_mapping(annotation),
                name=annotation.name,
                classifications=get_classifications(
                    annotation.classifications, []
                ),
            )
    return list(existing_tools.values())


def tool_mapping(
    annotation,
) -> Union[Mask, Polygon, Point, Rectangle, Line, TextEntity]:
    tool_types = ontology.Tool.Type
    mapping = {
        Mask: tool_types.SEGMENTATION,
        Polygon: tool_types.POLYGON,
        Point: tool_types.POINT,
        Rectangle: tool_types.BBOX,
        Line: tool_types.LINE,
        TextEntity: tool_types.NER,
    }
    result = mapping.get(type(annotation.value))
    if result is None:
        raise TypeError(
            f"Unexpected type found. {type(annotation.value)}. Expected one of {list(mapping.keys())}"
        )
    return result


def classification_mapping(annotation) -> Union[Text, Checklist, Radio]:
    classification_types = ontology.Classification.Type
    mapping = {
        Text: classification_types.TEXT,
        Checklist: classification_types.CHECKLIST,
        Radio: classification_types.RADIO,
    }
    result = mapping.get(type(annotation.value))
    if result is None:
        raise TypeError(
            f"Unexpected type found. {type(annotation.value)}. Expected one of {list(mapping.keys())}"
        )
    return result



import json
from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.mixins import CustomMetric
import labelbox.types as lb_types
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import (
    Label,
    ObjectAnnotation,
    RectangleUnit,
    Point,
    DocumentRectangle,
    DocumentEntity,
    DocumentTextSelection,
)

bbox_annotation = lb_types.ObjectAnnotation(
    name="bounding_box",  # must match your ontology feature's name
    value=lb_types.DocumentRectangle(
        start=lb_types.Point(x=42.799, y=86.498),  # Top left
        end=lb_types.Point(x=141.911, y=303.195),  # Bottom right
        page=1,
        unit=lb_types.RectangleUnit.POINTS,
    ),
)
bbox_labels = [
    lb_types.Label(
        data=lb_types.GenericDataRowData(global_key="test-global-key"),
        annotations=[bbox_annotation],
    )
]
bbox_ndjson = [
    {
        "bbox": {
            "height": 216.697,
            "left": 42.799,
            "top": 86.498,
            "width": 99.112,
        },
        "classifications": [],
        "dataRow": {"globalKey": "test-global-key"},
        "name": "bounding_box",
        "page": 1,
        "unit": "POINTS",
    }
]


def round_dict(data):
    if isinstance(data, dict):
        for key in data:
            if isinstance(data[key], (int, float)):
                data[key] = int(data[key])
            elif isinstance(data[key], dict):
                data[key] = round_dict(data[key])
            elif isinstance(data[key], (list, tuple)):
                data[key] = [round_dict(r) for r in data[key]]

    return data


def test_pdf():
    """
    Tests a pdf file with bbox annotations only
    """
    with open("tests/data/assets/ndjson/pdf_import.json", "r") as f:
        data = json.load(f)
    labels = [
        Label(
            uid=None,
            data=GenericDataRowData(
                uid="cl6xnv9h61fv0085yhtoq06ht",
            ),
            annotations=[
                ObjectAnnotation(
                    custom_metrics=[
                        CustomMetric(name="customMetric1", value=0.5),
                        CustomMetric(name="customMetric2", value=0.3),
                    ],
                    confidence=0.53,
                    name="boxy",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "5ad9c52f-058d-49c8-a749-3f20b84f8cd4",
                    },
                    value=DocumentRectangle(
                        start=Point(x=32.45, y=162.73),
                        end=Point(x=134.11, y=550.9),
                        page=4,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    name="boxy",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "20eeef88-0294-49b4-a815-86588476bc6f",
                    },
                    value=DocumentRectangle(
                        start=Point(x=251.42, y=223.26),
                        end=Point(x=438.2, y=680.3),
                        page=7,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    custom_metrics=[
                        CustomMetric(name="customMetric1", value=0.5),
                        CustomMetric(name="customMetric2", value=0.3),
                    ],
                    confidence=0.99,
                    name="boxy",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "641a8944-3938-409c-b4eb-dea354ed06e5",
                    },
                    value=DocumentRectangle(
                        start=Point(x=218.17, y=32.52),
                        end=Point(x=328.73, y=264.25),
                        page=6,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    custom_metrics=[
                        CustomMetric(name="customMetric1", value=0.5),
                        CustomMetric(name="customMetric2", value=0.3),
                    ],
                    confidence=0.89,
                    name="boxy",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "ebe4da7d-08b3-480a-8d15-26552b7f011c",
                    },
                    value=DocumentRectangle(
                        start=Point(x=4.25, y=117.39),
                        end=Point(x=169.08, y=574.3100000000001),
                        page=7,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    name="boxy",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "35c41855-575f-42cc-a2f9-1f06237e9b63",
                    },
                    value=DocumentRectangle(
                        start=Point(x=217.28, y=82.13),
                        end=Point(x=299.71000000000004, y=361.89),
                        page=8,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    name="boxy",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "1b009654-bc17-42a2-8a71-160e7808c403",
                    },
                    value=DocumentRectangle(
                        start=Point(x=83.34, y=298.12),
                        end=Point(x=83.72, y=501.95000000000005),
                        page=3,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
            ],
        ),
        Label(
            data=GenericDataRowData(
                uid="ckrb1sf1i1g7i0ybcdc6oc8ct",
            ),
            annotations=[
                ObjectAnnotation(
                    name="named_entity",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
                    },
                    value=DocumentEntity(
                        text_selections=[
                            DocumentTextSelection(
                                token_ids=[
                                    "3f984bf3-1d61-44f5-b59a-9658a2e3440f",
                                    "3bf00b56-ff12-4e52-8cc1-08dbddb3c3b8",
                                    "6e1c3420-d4b7-4c5a-8fd6-ead43bf73d80",
                                    "87a43d32-af76-4a1d-b262-5c5f4d5ace3a",
                                    "e8606e8a-dfd9-4c49-a635-ad5c879c75d0",
                                    "67c7c19e-4654-425d-bf17-2adb8cf02c30",
                                    "149c5e80-3e07-49a7-ab2d-29ddfe6a38fa",
                                    "b0e94071-2187-461e-8e76-96c58738a52c",
                                ],
                                group_id="2f4336f4-a07e-4e0a-a9e1-5629b03b719b",
                                page=1,
                            )
                        ]
                    ),
                )
            ],
        ),
    ]

    res = list(NDJsonConverter.serialize(labels))
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]


def test_pdf_with_name_only():
    """
    Tests a pdf file with bbox annotations only
    """
    with open("tests/data/assets/ndjson/pdf_import_name_only.json", "r") as f:
        data = json.load(f)

    labels = [
        Label(
            data=GenericDataRowData(
                uid="cl6xnv9h61fv0085yhtoq06ht",
            ),
            annotations=[
                ObjectAnnotation(
                    custom_metrics=[
                        CustomMetric(name="customMetric1", value=0.5),
                        CustomMetric(name="customMetric2", value=0.3),
                    ],
                    confidence=0.99,
                    name="boxy",
                    feature_schema_id=None,
                    extra={
                        "uuid": "5ad9c52f-058d-49c8-a749-3f20b84f8cd4",
                    },
                    value=DocumentRectangle(
                        start=Point(x=32.45, y=162.73),
                        end=Point(x=134.11, y=550.9),
                        page=4,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    name="boxy",
                    extra={
                        "uuid": "20eeef88-0294-49b4-a815-86588476bc6f",
                    },
                    value=DocumentRectangle(
                        start=Point(x=251.42, y=223.26),
                        end=Point(x=438.2, y=680.3),
                        page=7,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    name="boxy",
                    extra={
                        "uuid": "641a8944-3938-409c-b4eb-dea354ed06e5",
                    },
                    value=DocumentRectangle(
                        start=Point(x=218.17, y=32.52),
                        end=Point(x=328.73, y=264.25),
                        page=6,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    custom_metrics=[
                        CustomMetric(name="customMetric1", value=0.5),
                        CustomMetric(name="customMetric2", value=0.3),
                    ],
                    confidence=0.74,
                    name="boxy",
                    extra={
                        "uuid": "ebe4da7d-08b3-480a-8d15-26552b7f011c",
                    },
                    value=DocumentRectangle(
                        start=Point(x=4.25, y=117.39),
                        end=Point(x=169.08, y=574.3100000000001),
                        page=7,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    name="boxy",
                    extra={
                        "uuid": "35c41855-575f-42cc-a2f9-1f06237e9b63",
                    },
                    value=DocumentRectangle(
                        start=Point(x=217.28, y=82.13),
                        end=Point(x=299.71000000000004, y=361.89),
                        page=8,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
                ObjectAnnotation(
                    name="boxy",
                    extra={
                        "uuid": "1b009654-bc17-42a2-8a71-160e7808c403",
                    },
                    value=DocumentRectangle(
                        start=Point(x=83.34, y=298.12),
                        end=Point(x=83.72, y=501.95000000000005),
                        page=3,
                        unit=RectangleUnit.POINTS,
                    ),
                ),
            ],
        ),
        Label(
            data=GenericDataRowData(
                uid="ckrb1sf1i1g7i0ybcdc6oc8ct",
            ),
            annotations=[
                ObjectAnnotation(
                    name="named_entity",
                    extra={
                        "uuid": "f6879f59-d2b5-49c2-aceb-d9e8dc478673",
                    },
                    value=DocumentEntity(
                        text_selections=[
                            DocumentTextSelection(
                                token_ids=[
                                    "3f984bf3-1d61-44f5-b59a-9658a2e3440f",
                                    "3bf00b56-ff12-4e52-8cc1-08dbddb3c3b8",
                                    "6e1c3420-d4b7-4c5a-8fd6-ead43bf73d80",
                                    "87a43d32-af76-4a1d-b262-5c5f4d5ace3a",
                                    "e8606e8a-dfd9-4c49-a635-ad5c879c75d0",
                                    "67c7c19e-4654-425d-bf17-2adb8cf02c30",
                                    "149c5e80-3e07-49a7-ab2d-29ddfe6a38fa",
                                    "b0e94071-2187-461e-8e76-96c58738a52c",
                                ],
                                group_id="2f4336f4-a07e-4e0a-a9e1-5629b03b719b",
                                page=1,
                            )
                        ]
                    ),
                )
            ],
        ),
    ]
    res = list(NDJsonConverter.serialize(labels))
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]


def test_pdf_bbox_serialize():
    serialized = list(NDJsonConverter.serialize(bbox_labels))
    serialized[0].pop("uuid")
    assert serialized == bbox_ndjson



from typing import Any, Dict, Optional
from pydantic import ConfigDict
from labelbox.utils import _CamelCaseMixin


class App(_CamelCaseMixin):
    id: Optional[str] = None
    model_id: str
    name: str
    description: Optional[str] = None
    inference_params: Dict[str, Any]
    class_to_schema_id: Dict[str, str]
    ontology_id: str
    created_by: Optional[str] = None

    model_config = ConfigDict(protected_namespaces=())

    @classmethod
    def type_name(cls):
        return "App"


APP_FIELD_NAMES = list(App.model_json_schema()["properties"].keys())



from enum import Enum


class ConflictResolutionStrategy(str, Enum):
    KeepExisting = "KEEP_EXISTING"
    OverrideWithAnnotations = "OVERRIDE_WITH_ANNOTATIONS"
    OverrideWithPredictions = "OVERRIDE_WITH_PREDICTIONS"

    @staticmethod
    def from_str(label: str) -> "ConflictResolutionStrategy":
        return ConflictResolutionStrategy[label]
